"""
Perhaps better named "single" instead, this returns exactly one piece of feedback.


# Scoring

Some more notes on scoring.

NEGATIVE feedback function that IS TRIGGERED will subtract it's percentage.
    i.e., a bad thing happened
NEGATIVE feedback function that is NOT triggered will not subtract it's percentage.
POSITIVE feedback function that IS triggered will add it's percentage.
POSITIVE feedback function that is NOT triggered will not add it's percentage.

What happens when you have no `score` given? It's just a None. Probably should
    default to doing nothing, right? Just gets skipped in calculating the score.


Valence  | Triggered? | Operator | Result | Example
---------------------------------------------------
Positive | Yes        | +        | Add 5  | compliment()
Negative | Yes        | +        | 0      | assert_equal()
Positive | No         | +        | 0      | compliment()
Negative | No         | +        | Add 5  | assert_equal()
Positive | Yes        | -        | 0      | compliment()
Negative | Yes        | -        | Sub 5  | assert_equal()
Positive | No         | -        | Sub 5  | compliment() * Weird situation
Negative | No         | -        | 0      | assert_equal()

"+N": Default behavior for a float argument
"+N%": allows you to use whole numbers ("+20%" == "+.2" == .2)
"+N/Z": allows you to specify the total number of points consistently.
"-N"
"-N%"
"*N"
"*N%"
"/N"
"/N%"
"=N": set the score absolutely
"=N%": set the score to the percentage
"^N" or "^N%": Cap the maximum number of points possible
"_N" or "_N%": Cap the minimum number of points possible

How do scores work within a group? How do you specify that point allocations
happen within a group, or absolutely overall?
* Unit test - we control, this so teacher API is for "this function overall"
* Sections - teachers write statement level stuff
* Phase - teachers write statement level stuff

So I think scores default to their current group OR global depending on context.
You can override a score given within a group to be global using "global"?
    "+N% global"

You could also have "scoring modes" to change default behavior.
    { (Feedback.NEGATIVE, None): "=0%" }
    set_score_mode("Negative None is =0%")
        If you encounter a Negative None, then the score gets set to 0%.

I'd like to be able to use constants too, I think.
    set_grading_constant("UNITTESTS", .25)
    "+UNITTESTS%"
    "+DEFINITION%"

```python
# Give 5% if this does not get triggered
assert_equal(4, 5, score="+5%")

# Give 5% if this does get triggered
give_partial("+5%")
```

"""
from pedal.core.report import MAIN_REPORT
from pedal.core.final_feedback import FinalFeedback, set_correct_no_errors
from pedal.core.feedback import Feedback, DEFAULT_CATEGORY_PRIORITY
from pedal.resolvers.core import make_resolver


def by_priority(feedback):
    """
    Converts a feedback into a numeric representation for sorting.

    Args:
        feedback (Feedback): The feedback object to convert
    Returns:
        float: A decimal number representing the feedback's relative priority.
    """
    category = Feedback.CATEGORIES.UNKNOWN
    if feedback.category is not None:
        category = feedback.category.lower()
    priority = 'medium'
    if feedback.priority is not None:
        priority = feedback.priority.lower()
        priority = Feedback.CATEGORIES.ALIASES.get(priority, priority)
    if category in DEFAULT_CATEGORY_PRIORITY:
        value = DEFAULT_CATEGORY_PRIORITY.index(category)
    else:
        value = len(DEFAULT_CATEGORY_PRIORITY)
    if priority in DEFAULT_CATEGORY_PRIORITY:
        value = DEFAULT_CATEGORY_PRIORITY.index(priority)
        priority = 'medium'
    offset = priority_offset(priority)
    return value + offset


def priority_offset(priority):
    """

    Args:
        priority:

    Returns:

    """
    if priority == 'low':
        return .7
    elif priority == 'medium':
        return .5
    elif priority == 'high':
        return .3
    else:
        return .1


@make_resolver
def resolve(report=MAIN_REPORT, priority_key=by_priority):
    """
    Args:
        priority_key: The key function to sort feedbacks by
        report (Report): The report object to resolve down. Defaults to the
                         global MAIN_REPORT

    Returns
        str: A string of HTML feedback to be delivered
    """
    # Prepare feedbacks
    report.finalize_feedbacks()
    feedbacks = report.feedback + report.ignored_feedback
    feedbacks.sort(key=priority_key)
    # Create the initial final feedback
    final = set_correct_no_errors(report)
    # Process each feedback in turn
    for feedback in feedbacks:
        final.merge(feedback)
    # Override empty message
    final.finalize()
    report.result = final
    report.resolves.append(final)
    return final




from pedal.assertions.feedbacks import AssertionFeedback
from pedal.core.report import MAIN_REPORT


class missing_dataclass(AssertionFeedback):
    """ Unconditionally asserts that the dataclass is missing. """
    title = "Missing Dataclass"
    message_template = "No dataclass named {name_message} was found."

    def __init__(self, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        super().__init__(fields=fields, **kwargs)


class duplicate_dataclass_definition(AssertionFeedback):
    """ Unconditionally assert that the dataclass is redefined somewhere. """
    title = "Duplicate Dataclass Definition"
    message_template = ("The dataclass {name_message} was defined multiple times, "
                        "on lines {lines_message}.")

    def __init__(self, name, lines, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        lines_message = ", ".join(str(line) for line in lines[:-1])
        lines_message += " and " + str(lines[-1])
        fields['lines'] = lines
        fields['lines_message'] = lines
        super().__init__(fields=fields, **kwargs)


class too_few_fields(AssertionFeedback):
    """ Too Few Fields """
    title = "Too Few Fields"
    message_template = ("The dataclass named {name_message} has fewer "
                        "fields ({found}) than expected ({expected}).")

    def __init__(self, name, found, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['found'] = found
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class too_many_fields(AssertionFeedback):
    """ Too Many Fields """
    title = "Too Many Fields"
    message_template = ("The dataclass named {name_message} has more "
                        "fields ({found}) than expected ({expected}).")

    def __init__(self, name, found, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['found'] = found
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class invalid_field_type(AssertionFeedback):
    """ Invalid field type """
    title = "Invalid Field Type"
    message_template = ("The dataclass named {name_message} has a field "
                        "named {field_message} with an invalid type.")

    def __init__(self, name, field, invalid_type, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['field'] = field
        fields['field_message'] = report.format.name(field)
        fields['invalid_type'] = invalid_type
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class missing_field_type(AssertionFeedback):
    """ Invalid field type """
    title = "Missing Field Type"
    message_template = ("The dataclass named {name_message} has a field "
                        "named {field_message}, but that field does "
                        "not have a type specified.")

    def __init__(self, name, field, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['field'] = field
        fields['field_message'] = report.format.name(field)
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class unknown_field(AssertionFeedback):
    """ Unknown Field """
    title = "Unknown Field"
    message_template = ("The dataclass named {name_message} had a field named {field_message} "
                        "but that field is not supposed to be there. Are you sure you got "
                        "the name right?")

    def __init__(self, name, field, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['field'] = field
        fields['field_message'] = report.format.name(field)
        super().__init__(fields=fields, **kwargs)


class wrong_fields_type(AssertionFeedback):
    """ Wrong Fields Type """
    title = "Wrong Fields Type"
    message_template = ("The dataclass named {name_message} has a field "
                        "named {field_message} that is {actual_message},"
                        " but should be {expected_message}.")

    def __init__(self, name, field, actual, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['field'] = field
        fields['field_message'] = report.format.name(field)
        fields['actual'] = actual
        fields['actual_message'] = actual.precise_description()
        fields['expected'] = expected
        fields['expected_message'] = expected.precise_description()
        super().__init__(fields=fields, **kwargs)


class wrong_field_order(AssertionFeedback):
    """ Wrong Field Order """
    title = "Wrong Field Order"
    message_template = ("The dataclass named {name_message} has a field named "
                        "{actual_message}, which is correct. However, that field should "
                        "come after the field named {expected_message}. Instead, it came "
                        "before. Check the order of your definitions' fields.")

    def __init__(self, name, actual, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['actual'] = actual
        fields['actual_message'] = report.format.name(actual)
        fields['expected'] = expected
        fields['expected_message'] = report.format.name(expected)
        super().__init__(fields=fields, **kwargs)


class name_is_not_a_dataclass(AssertionFeedback):
    """ The name is not a dataclass """
    title = "Name Is Not a Dataclass"
    message_template = "You defined `{name_message}`, but did not define it as a dataclass."

    def __init__(self, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        super().__init__(fields=fields, **kwargs)


class dataclass_not_available(AssertionFeedback):
    """ The name is not a dataclass """
    title = "Dataclass Not Available"
    message_template = ("You may have defined `{name_message}`, but it was not available"
                        " to be called in the top-level scope. Perhaps you mistakenly defined"
                        " it inside another dataclass or scope?")

    def __init__(self, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        super().__init__(fields=fields, **kwargs)


class missing_dataclass_annotation(AssertionFeedback):
    """ The dataclass is missing the annotation """
    title = "Dataclass Annotation Missing"
    message_template = ("You have defined {name_message}, but you did not include"
                        " the {dc_annotation} annotation. Make sure you import and include"
                        " the {dc_annotation} annotation directly before the class definition.")

    def __init__(self, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['dc_annotation'] = report.format.html_code('@dataclass')
        super().__init__(fields=fields, **kwargs)



"""
List of features to test:

* Setting context_message and assertion_message to False
* Setting context_message and assertion_message to strings
* Setting explanation to be a non-empty string

* Testing at top-level
* Testing within a function
* Testing within a subtest

* Testing call() on both sides
* Testing call() on either side
* Testing no call()

Two sandboxed results:
    The value of `{first_name}` was:\n<pre>{first_value}</pre>
    The value of `{second_name}` was:\n<pre>{second_value}</pre>
    But I expected `{first_name}` to be equal to `{second_name}`.
One sandboxed result:
    The value of `{first_name}` was:\n<pre>{first_value}</pre>
    But I expected `{first_name}` to be equal to:\n<pre>{expected_value}</pre>
No sandbox result:
    <pre>{first_value} != {expected_value}</pre>



assert_not_equal(call('add', 5, 4), 1,
                 context = None, assertion = None,
                 explanation="Your add function is subtracting.",
                 label="test_add_but_subtracted_5_4")

assert_equal(call('add', 5, 4), 9,
             explanation="You added incorrectly."
             label="test_add_5_4")


"""

from pedal.core.feedback import Feedback, FeedbackGroup
from pedal.core.feedback_category import FeedbackStatus
from pedal.core.report import MAIN_REPORT
from pedal.sandbox import Sandbox
from pedal.sandbox.data import format_contexts
from pedal.sandbox.exceptions import SandboxException
from pedal.sandbox.result import is_sandbox_result, unwrap_value
from pedal.assertions.constants import TOOL_NAME
from pedal.utilities.text import chomp


class InterpolatedValue:
    """ Abstract wrapper around asserted values to handle different ways
    of printing them. """

    def __init__(self, value):
        self.value = value
        self.is_sandboxed = is_sandbox_result(value)
        if isinstance(value, Exception):
            self.is_error = True
        # Sandboxes with exceptions become their exception
        elif isinstance(value, Sandbox) and value.exception:
            self.is_error = True
            self.value = value.exception
            self.is_sandboxed = False
        else:
            self.is_error = False
        self.report = MAIN_REPORT
        if self.is_sandboxed:
            context_id = value._actual_context_id
            sandbox = value._actual_sandbox
            self.context = sandbox.get_context(context_id)
        else:
            self.context = None

    def set_report(self, report):
        """ Update the report on this value. """
        self.report = report


class ExactValue(InterpolatedValue):
    """ Wrapper around literal values to produce them unmodified. """

    def __str__(self):
        return " "+str(self.value)


class SandboxedValue(InterpolatedValue):
    """ Wrapper around sandboxed values to preformat their text. """

    def __str__(self):
        return ":\n"+self.report.format.python_value(repr(self.value))


class AssertionBreak(Exception):
    """
    Exception that can be raised to exit out of an assertion, jumping control
    flow around subsequent checks.
    """

    def __init__(self, feedback, exception=None):
        self.feedback = feedback
        self.exception = exception

    def __str__(self):
        exception_name = (self.exception.__class__.__name__
                          if self.exception is not None else "No Error")
        return f"AssertionBreak({self.feedback.label}, {exception_name})"


class AssertionFeedback(Feedback):
    """ Most basic assertion """
    title = "Failed Instructor Test"
    category = Feedback.CATEGORIES.SPECIFICATION
    valence = Feedback.NEGATIVE_VALENCE
    kind = Feedback.KINDS.CONSTRAINT
    assertion_status: str


class RuntimeAssertionFeedback(AssertionFeedback):
    """ Class for representing constraints asserted by an instructor """
    message_template = ("Student code failed instructor test.\n"
                        "{context_message}"
                        "{assertion_message}"
                        "{explanation}")
    # TODO: The explanation field is broken, because it's not a keyword parameter
    _expected_verb: str
    _aggregate_verb = "Expected"
    _inverse_operator: str

    def __init__(self, left, right, *args, **kwargs):
        self.report = kwargs.get('report', MAIN_REPORT)
        left.set_report(self.report)
        right.set_report(self.report)
        # Get contexts
        contexts = self.get_sandbox_contexts([left, right])
        # Calculate the context_message
        if kwargs.get('context') is False:
            context_message = ""
        elif kwargs.get('context') is not None:
            context_message = kwargs['context']
        else:
            context_message = format_contexts(contexts, self.report.format)
        # Calculate the assertion_message
        if kwargs.get('assertion') is False:
            assertion_message = ""
        elif kwargs.get('assertion') is not None:
            assertion_message = kwargs['assertion'] + "\n"
        else:
            assertion_message = self.format_assertion(left, right, contexts)
        # Calculate explanation
        explanation = kwargs.get("explanation", "")
        # Add in new fields
        fields = kwargs.setdefault('fields', {})
        fields['left'] = left.value
        fields['right'] = right.value
        fields['left_boxed'] = left
        fields['right_boxed'] = right
        fields['contexts'] = contexts
        fields['expected_verb'] = self._expected_verb
        fields['aggregate_verb'] = self._aggregate_verb
        fields['inverse_operator'] = self._inverse_operator
        fields['context_message'] = context_message
        fields['assertion_message'] = assertion_message
        fields['explanation'] = explanation

        try:
            super().__init__(left, right, *args, **kwargs)
        except Exception as e:
            #if not hasattr(self, "_met_condition"):
            #    self._met_condition = False
            #    raise e
            parent = self.report.get_current_group()
            # TODO: Does this handle nested groups correctly?
            if parent is not None:
                if not parent.try_all:
                    raise AssertionBreak(self, e)
        if not self:
            if self.report[TOOL_NAME]['exceptions']:
                raise AssertionBreak(self)

    def get_sandbox_contexts(self, wrapped_values):
        """ Retrieve any sandbox contexts associated with these values. """
        contexts = []
        for wrapped_value in wrapped_values:
            if wrapped_value.is_sandboxed:
                contexts.append(wrapped_value.context)
            if isinstance(wrapped_value.value, Sandbox):
                run_contexts = wrapped_value.value.get_context()
                if run_contexts:
                    contexts.append(run_contexts)
        return contexts

    def _build_result_from_target(self, contexts, index):
        target = contexts[index][-1].target
        if target == "_" or target is None:
            if len(contexts) == 1:
                return "the result"
            elif index == 0:
                return "the first result"
            else:
                return "the second result"
        return self.report.format.name(target)

    def format_assertion(self, left, right, contexts):
        """
        Creates a textual explanation of the difference between left and right,
        using any targets from the contexts. The ``expected_verb`` is the
        human-friendly description of the comparison being made between the two
        sides, while ``inverse_operator`` is the actual comparison that was
        performed.

        Args:
            left (InterpolatedValue):
            right (InterpolatedValue):
            contexts (list[:py:class:`pedal.sandbox.data.SandboxContext`]):
        Returns:

        """
        # Handle whether either side had an error
        if left.is_error or right.is_error:
            # TODO:
            assertion = self.format_exception(left, right)
        # Handle the number of contexts
        elif not contexts:
            # TODO: Check if this is working correctly; might be wrapping in output weirdly
            assertion = self.report.format.output(f"{left.value} "
                                                  f"{self._inverse_operator} "
                                                  f"{right.value}")
        elif len(contexts) == 1:
            # If the expected_verb is a tuple, the right side's value is used
            #   to determine which of the two possible messages should be used.
            #   If the right side is a sandboxed value, uses the second; else the first.
            _expected_verb = self._expected_verb
            if isinstance(_expected_verb, tuple):
                relevant_index = int(right.is_sandboxed)
                _expected_verb = _expected_verb[relevant_index]
            # Make sure ``left`` is the sandbox result
            if right.is_sandboxed:
                left, right = right, left
            target = self._build_result_from_target(contexts, 0)
            assertion = (f"The value of {target} was{left}\n"
                         f"But I expected {target} {_expected_verb}{right}")
        elif len(contexts) == 2:
            first_target = self._build_result_from_target(contexts, 0)
            second_target = self._build_result_from_target(contexts, 1)
            assertion = (f"The value of {first_target} was{left}\n"
                         f"The value of {second_target} was{right}\n"
                         f"But I expected {first_target} {self._expected_verb}{second_target}")
        else:
            # Invalid state, why'd you have 3+ contexts?
            assertion = ""
        return assertion

    def format_exception(self, left, right):
        """ Create a simple formatted exception message """
        assertion = "The following exception occurred:\n"
        if left.is_error:
            if not isinstance(left.value, SandboxException):
                assertion += unwrap_value(left.value.feedback._get_message())
            else:
                assertion += str(left.value)
            #assertion += self.report.format.output(str(left.value))
            self.suppress_runtime_error(left.value)
        if right.is_error:
            if not isinstance(right.value, SandboxException):
                assertion += unwrap_value(right.value.feedback._get_message())
            else:
                assertion += str(right.value)
            #assertion += self.report.format.output(str(right.value))
            self.suppress_runtime_error(right.value)
        return assertion

    def suppress_runtime_error(self, exception):
        if hasattr(exception, "feedback") and exception.feedback is not None:
            exception.feedback.parent = self


class RuntimePrintingAssertionFeedback(RuntimeAssertionFeedback):
    """ Variant for handling printing instead of return value"""

    def __init__(self, execution, text, exact_strings=False, **kwargs):
        super().__init__(SandboxedValue(execution), SandboxedValue(text),
                         exact_strings=exact_strings, **kwargs)

    def format_assertion(self, left, right, contexts):
        """
        Overrides the regular format_assertion to write better descriptions
        of the text.
        """
        # Handle whether either side had an error
        if left.is_error or right.is_error:
            assertion = self.format_exception(left, right)
            return assertion
        # Sandbox
        if isinstance(left.value, Sandbox):
            actual_output = chomp(left.value.raw_output)
            if not actual_output:
                actual = "There was no printed output."
                actual_output = ""
            else:
                actual = "The printed output was:"
                actual_output = self.report.format.output(actual_output)
        # Sandboxed value
        else:
            actual_output = chomp(left.context[-1].output)
            if not actual_output:
                actual = "The function did not print."
                actual_output = ""
            else:
                actual = "The function printed:"
                actual_output = self.report.format.output(actual_output)
        # Get the expected
        expected = f"But I expected {self._expected_verb}:"
        expected_output = self.report.format.output(right.value)
        # Join everything
        return "\n".join((actual, actual_output, expected, expected_output))

    def get_output(self, execution):
        """ Get the actual output from the execution"""
        if isinstance(execution.value, Sandbox):
            return chomp(execution.value.raw_output)
        # Sandboxed value
        else:
            return chomp(execution.context[-1].output)


"""
I ran your function `add` on some new arguments.
{I also entered some inputs of my own.}?
It passed X/Y tests.

|    | Arguments | Returned | Expected |
|----|-----------|----------|----------|
| ❌ |  1, 2     |   -1     |     3    |


What if they trigger an exception instead of producing a value? 

What if they assert_prints?
    Could be a separate table. Perhaps all the asserts get their own table?
    Yes, this would make a lot of sense. Then each table is adjusting their
        "Expected" column title as appropriate.
    And some tables want to also adjust their Returned column to be "Printed".
        And to also include the Input column, I suppose (although maybe
            that should happen automatically whenever you include inputs).
        Or perhaps it always tells you explicitly what was printed (if it's available?)
            I feel like that would be information overload.

    "Expected the output to be"

If you encounter an existing AssertionGroup feedback, concatenate it.

The final result is turned into a new AssertionFeedback that is placed before
    its composite children, to ensure that is given priority. Perhaps we should
    also mark certain feedback as "superseded" by others?

Group Manager:
    Keep track of current groups in a stack
    Groups have a name
    Also record
        Errors, failures, successes
            of each AssertionFeedback that occurs
    Two modes:
        Try all: Keep going even if you encounter an error or a failure.
        Fail on first: If you encounter any fails/errors, stop checking.
        
Automatically mutes/unscores any collected assertion feedbacks.

"""


class assert_group(AssertionFeedback, FeedbackGroup):
    """

    TODO: Decorator version

    """

    message_template = ("Student code failed instructor tests.\n"
                        "{summary_statistics}\n"
                        "{extra_context}"
                        "{tables}")

    def __init__(self, name, try_all=True, context=None, **kwargs):
        super().__init__(delay_condition=True, **kwargs)
        self.name = name
        self.try_all = try_all
        self.successes = []
        self.failures = []
        self.errors = []
        self.all_feedback = []

        self.fields['extra_context'] = ""
        if context:
            formatted_messages = format_contexts([context], self.report.format)
            if formatted_messages.strip():
                self.fields['extra_context'] = formatted_messages + '\n'

    def __enter__(self):
        self.report.start_group(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Go through all my collected assertions
        # Count failures, errors, successes
        # If there are any non-successes, then produce a summary table.
        #   Otherwise produce a success within this group
        # Mute and unscore any feedbacks (use their scores though)
        self.report.stop_group(self)
        self.format_message()
        self.format_table()
        self._handle_condition()
        return False

    def _get_child_feedback(self, feedback, active):
        if isinstance(feedback, RuntimeAssertionFeedback):
            if feedback._status == FeedbackStatus.INACTIVE:
                self.successes.append(feedback)
            elif feedback._status == FeedbackStatus.ACTIVE:
                self.failures.append(feedback)
            elif feedback._status == FeedbackStatus.ERROR:
                self.errors.append(feedback)
            self.all_feedback.append(feedback)
        feedback.muted = True
        feedback.unscored = True

    def format_message(self):
        self.fields['failures'] = self.failures
        self.fields['errors'] = self.errors
        self.fields['successes'] = self.successes
        self.fields['failure_count'] = failure_count = len(self.failures)
        self.fields['error_count'] = error_count = len(self.errors)
        self.fields['success_count'] = success_count = len(self.successes)
        self.fields['total_count'] = total_count = len(self.all_feedback)
        self.fields['name'] = self.name
        self.fields['try_all'] = self.try_all

        stats = [f"You passed {success_count}/{total_count} tests.\n"]
        if error_count:
            stats.append(f"There were {error_count} errors.\n")
        self.fields['summary_statistics'] = "\n".join(stats)

    def format_table(self):
        # Group by their tables
        # TODO: left target, right target, expected_verb
        groups = {}
        for feedback in self.all_feedback:
            left = feedback.fields['left_boxed']
            right = feedback.fields['right_boxed']
            verb = feedback.fields['aggregate_verb']
            left_called = left.context[-1].called if left.context else None
            right_called = right.context[-1].called if right.context else None
            key = (left_called, right_called, verb)
            if key not in groups:
                groups[key] = []
            groups[key].append(feedback)

        # Actually build tables
        tables = []
        for (left_called, right_called, verb), feedbacks in groups.items():
            if left_called is None and right_called is None:
                continue
            # TODO: Handle left AND right called
            called = self.report.format.name(left_called or right_called)
            # TODO: Handle inputs, outputs
            columns = ["", "Arguments", "Returned", verb]
            rows = []
            for feedback in feedbacks:
                left = feedback.fields['left_boxed']
                right = feedback.fields['right_boxed']
                if left.context:
                    arguments = left.context[-1].args
                    actual, expected = str(left.value), str(right.value)
                else:
                    arguments = right.context[-1].args
                    actual, expected = str(right.value), str(left.value)
                if feedback._status == FeedbackStatus.INACTIVE:
                    outcome = self.report.format.check_mark()
                else:
                    outcome = self.report.format.negative_mark()
                rows.append([outcome, self.report.format.python_code(arguments), actual, expected])
            table = self.report.format.table(rows, columns)
            table = (f"I ran your function {called} on some new arguments."
                     f"{table}")
            tables.append(table)
        self.fields['tables'] = "\n".join(tables)

    def condition(self):
        """ Check that there are no errors or failures. """
        return self.errors or self.failures


"""
# Group a bunch of tests at the top level
with assert_group('add') as add_group:
    assert_equal(call('add', 1, 2), 3)
    assert_equal(call('add', 4, 5), 9)
    assert_equal(call('add', -4, -2), -6)

if add_group:
    pass


# Parameterize the tests
@assert_group('add')
def test_add(first_arg):
    assert_equal(call('add', first_arg, 2), first_arg + 3)
    assert_equal(call('add', first_arg, 5), first_arg + 9)
    assert_equal(call('add', first_arg, -2), first_arg + -6)
    assert_greater(call('add', first_arg, 3), 0)


if test_add(5) and test_add(-3):
    set_correct()


def unit_test(function_name, *tests):
    with assert_group(function_name) as group:
        for test in tests:
            args, expected = test
            assert_equal(call(function_name, *args), expected)
    return group


if unit_test('add', [[1, 2], 3]):
    pass
"""



import sys
import os

from pedal import verify, allow_real_io, block_real_io, get_python_errors
from pedal.core.environment import Environment
from pedal.core.report import MAIN_REPORT
from pedal.core.commands import set_correct as set_correct_feedback
from pedal.resolvers.core import make_resolver
from pedal.sandbox import run, get_sandbox, set_input, start_trace
from pedal.sandbox.feedbacks import runtime_error
from pedal.source.feedbacks import syntax_error
from pedal.tifa import tifa_analysis
from pedal.resolvers.simple import resolve
from pedal.core.formatting import Formatter

try:
    import tabulate
except ImportError:
    tabulate = None

RESET = "\033[0;0m"
REVERSE = "\033[;7m"
UNDERLINE = "\033[4m"
# \033[1m\033[96m
BOLD_BRIGHT_COLOR = "\033[1;96m"
BOLD_RED = "\033[1;91m"
BOLD_GREEN = "\033[1;92m"
CHECKMARK = "\u2713"
CROSSMARK = "\u2717"


class TerminalFormatter(Formatter):
    def __init__(self, report=MAIN_REPORT, path_mask=None, **kwargs):
        super().__init__(report, **kwargs)
        self.path_mask = path_mask

    def name(self, name):
        return f"{BOLD_BRIGHT_COLOR}{name}{RESET}"

    def line(self, line_number):
        return f"{UNDERLINE}{line_number}{RESET}"

    def filename(self, filename):
        if self.path_mask:
            filename = filename.replace(self.path_mask, '')
        else:
            filename = os.path.basename(filename)
        return f"{UNDERLINE}{filename}{RESET}"

    def table(self, rows, columns):
        if tabulate:
            return "\n" + self.pre(tabulate.tabulate(rows, headers=columns, tablefmt="plain"))
        else:
            return super().table(rows, columns)


@make_resolver
def resolve_on_terminal(report=MAIN_REPORT):
    # print("\033[47m", "-" * 35, "FEEDBACK", "-" * 35, "\033[0m", file=file)
    feedback = resolve(report)
    print("")
    print(f"{REVERSE}FEEDBACK{RESET} Based on your code, here are some tips and recommendations:\n")
    if feedback.correct:
        print(f"{BOLD_GREEN}{CHECKMARK}{RESET}️  Your code ran successfully.\n")
        print(f"{feedback.message}")
    else:
        print(f"{BOLD_RED}{CROSSMARK}{RESET}  {feedback.title}\n")
        print(f"{feedback.message}")
    print("")

    # if feedback.category not in (FeedbackCategory.SYNTAX, FeedbackCategory.RUNTIME):
    # Print out the first runtime/syntax error that we encounter
    for python_error in get_python_errors(report):
        if python_error:
            python_error_native_message = to_native_message(python_error)
            if python_error_native_message:
                print(f"{REVERSE}TERMINAL OUTPUT{RESET} For reference, here is the original Python error:")
                print(python_error_native_message)
        break
    return feedback


def to_native_message(error):
    if not any(field in error.fields for field in ['traceback_message', 'exception_name', 'exception_message']):
        return ""
    return f"\n{error.fields.get('traceback_message', '')}\n{error.fields.get('exception_name', '')}: {error.fields.get('exception_message', '').strip()}"


def enhance_native_errors():
    syntax_error.override(message_template="Bad syntax on line {lineno:line}.\n{suggestion_message}")
    runtime_error.override(message_template=(
        "{context_message}\n"
        "{exception_name_proper} occurred:\n\n"
        "{exception_message}\n\n"
        "{suggestion_message}"
    ))


class TerminalEnvironment(Environment):
    """
    Configures the BlockPy programming environment.

    Args:
        main_file (str): If `files` and `main_code` are not provided, then this should be the path to a file
            that contains the student's code. If `files` is not given but `main_code` is, then `main_code` will be
            the contents of that file and `main_file` will be the filename (defaults to `answer.py`). If `files` is
            given, then `main_file` will be used to select a `files` from that dictionary (defaults to `answer.py`).
    """

    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='on_run.py', skip_tifa=False, skip_run=False,
                 inputs=None, set_correct=True, set_success=None,
                 report=MAIN_REPORT, trace=True, threaded=False, real_io=True, path_mask=None):
        super().__init__(files=files, main_file=main_file, main_code=main_code,
                         user=user, assignment=assignment, course=course,
                         execution=execution, instructor_file=instructor_file,
                         report=report)
        enhance_native_errors()
        report.set_formatter(TerminalFormatter(report, path_mask=path_mask))
        set_correct_feedback.override(message_template=("Great work! Based on your code submitted, there are no other "
                                                        "recommendations available. You can proceed to the next page "
                                                        "by using the “Next” button in your lesson."))
        verify(report=self.report)
        if not skip_tifa:
            tifa_analysis(report=self.report)
        if inputs:
            set_input(inputs)
        if skip_run:
            student = get_sandbox(report=report)
        else:
            print(f"{REVERSE}YOUR CODE{RESET} We ran your code. Here's the output:\n")
            if trace:
                start_trace()
            if real_io:
                allow_real_io()
            student = run(report=report, threaded=threaded)
            if real_io:
                block_real_io()
        student.threaded = threaded
        self.fields = {
            'student': student,
            'resolve': resolve_on_terminal
        }


setup_environment = TerminalEnvironment



from pedal.sandbox.mocked import MockModule


class MockPlt(MockModule):
    """
    Mock MatPlotLib library that can be used to capture plot data.

    Attributes:
        plots (list of dict): The internal list of plot dictionaries.
    """

    def __init__(self):
        super().__init__()
        self._reset_plots()

    def show(self, **kwargs):
        """ Renders the plot """
        self.plots.append(self.active_plot)
        self._reset_plot()

    def unshown_plots(self):
        """ Checks for plots that are not yet shown. """
        return self.active_plot['data']

    def __repr__(self):
        return repr(self.plots)

    def __str__(self):
        return str(self.plots)

    def _reset_plots(self):
        self.plots = []
        self._reset_plot()

    def _reset_plot(self):
        self.active_plot = {'data': [],
                            'xlabel': None, 'ylabel': None,
                            'title': None, 'legend': False,
                            'kwargs': {},
                            'xlabel_kwargs': {}, 'ylabel_kwargs': {},
                            'title_kwargs': {}, 'suptitle_kwargs': {},
                            'legend_kwargs': {},
                            'other_calls': []}

    def hist(self, data, **kwargs):
        """ Make a histogram """
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'hist', 'values': data,
                                         'label': label, 'kwargs': kwargs})

    def boxplot(self, data, **kwargs):
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'boxplot', 'values': data,
                                         'label': label, 'kwargs': kwargs})

    def hlines(self, y, xmin, xmax, **kwargs):
        label = kwargs.get('label')
        self.active_plot['data'].append({'type': 'hlines', 'y': y,
                                         'xmin': xmin, 'xmax': xmax,
                                         'label': label, 'kwargs': kwargs})

    def vlines(self, x, ymin, ymax, **kwargs):
        label = kwargs.get('label')
        self.active_plot['data'].append({'type': 'vlines', 'x': x,
                                         'ymin': ymin, 'ymax': ymax,
                                         'label': label, 'kwargs': kwargs})

    def plot(self, xs, ys=None, **kwargs):
        """ Make a line plot """
        label = kwargs.get('label', None)
        if ys is None:
            self.active_plot['data'].append({'type': 'line',
                                             'x': list(range(len(xs))),
                                             'y': xs, 'label': label,
                                             'kwargs': kwargs})
        else:
            self.active_plot['data'].append({'type': 'line', 'x': xs,
                                             'y': ys, 'label': label,
                                             'kwargs': kwargs})

    def scatter(self, xs, ys, **kwargs):
        """ Make a scatter plot """
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'scatter', 'x': xs,
                                         'y': ys, 'label': label,
                                         'kwargs': kwargs})

    def bar(self, xs, ys, **kwargs):
        """ Make a bar plot """
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'bar', 'x': xs,
                                         'y': ys, 'label': label,
                                         'kwargs': kwargs})

    def xlabel(self, label, **kwargs):
        """ Label the x-axis """
        self.active_plot['xlabel'] = label
        self.active_plot['xlabel_kwargs'].update(kwargs)

    def title(self, label, **kwargs):
        """ Make the title """
        self.active_plot['title'] = label
        self.active_plot['title_kwargs'].update(kwargs)

    def suptitle(self, label, **kwargs):
        """ Make the super title """
        self.title(label, **kwargs)
        self.active_plot['suptitle_kwargs'].update(kwargs)

    def ylabel(self, label, **kwargs):
        """ Label the Y-axis """
        self.active_plot['ylabel'] = label
        self.active_plot['ylabel_kwargs'].update(kwargs)

    def legend(self, **kwargs):
        """ Show the legend """
        self.active_plot['legend'] = True
        self.active_plot['legend_kwargs'].update(kwargs)

    def _generate_patches(self):
        def dummy(*args, **kwargs):
            """ This function does nothing. """
            # TODO: Capture name; does this need to be a decorated function?
            self.active_plot['other_calls'].append((args, kwargs))

        return dict(hist=self.hist, plot=self.plot,
                    bar=self.bar, boxplot=self.boxplot,
                    hlines=self.hlines, vlines=self.vlines,
                    scatter=self.scatter, show=self.show,
                    xlabel=self.xlabel, ylabel=self.ylabel,
                    title=self.title, legend=self.legend,
                    xticks=dummy, yticks=dummy,
                    autoscale=dummy, axhline=dummy,
                    axhspan=dummy, axvline=dummy,
                    axvspan=dummy, clf=dummy,
                    cla=dummy, close=dummy,
                    figlegend=dummy, figimage=dummy,
                    suptitle=self.suptitle, text=dummy,
                    tick_params=dummy, ticklabel_format=dummy,
                    tight_layout=dummy, xkcd=dummy,
                    xlim=dummy, ylim=dummy,
                    xscale=dummy, yscale=dummy)



import unittest
import os
import sys
from textwrap import dedent

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pedal.core import *
from pedal.core.commands import clear_report, get_all_feedback, contextualize_report
from pedal.source import *
from pedal.tifa import tifa_analysis
from tests.execution_helper import Execution
import pedal.resolvers.sectional as sectional
from pedal.utilities.system import IS_AT_LEAST_PYTHON_311, IS_AT_LEAST_PYTHON_39, IS_AT_LEAST_PYTHON_38, IS_AT_LEAST_PYTHON_310


class TestCode(unittest.TestCase):

    def test_catches_blank_files(self):
        contextualize_report('')
        verify()
        feedback = get_all_feedback()
        self.assertTrue(feedback)
        self.assertEqual(feedback[0].label, 'blank_source')

        contextualize_report('                  \n\n\n \n\n\n      ')
        verify()
        feedback = get_all_feedback()
        self.assertTrue(feedback)
        self.assertEqual(feedback[0].label, 'blank_source')

    def test_catches_syntax_errors(self):
        contextualize_report('a b c')
        verify()
        feedback = get_all_feedback()
        self.assertTrue(feedback)
        self.assertEqual(feedback[0].label, 'syntax_error')

    def test_syntax_traceback_good(self):
        with Execution('a=0\nb b b\nc = 0') as e:
            pass
        self.assertFalse(e.final.success)
        self.assertEqual(e.final.label, 'syntax_error')
        self.assertEqual("""Bad syntax on line 2.

The traceback was:
Line 2 of file answer.py
    b b b
""" + ('      ^\n' if IS_AT_LEAST_PYTHON_311 else "") + """
Invalid syntax.
Suggestion: Check line 2, the line before it, and the line after it. Ignore blank lines.""", e.final.message)

    def test_no_more_input(self):
        contextualize_report('def x():')
        verify()
        feedback = get_all_feedback()
        self.assertTrue(feedback)
        self.assertEqual(feedback[0].label, 'indentation_error' if IS_AT_LEAST_PYTHON_39 else "syntax_error")
        if IS_AT_LEAST_PYTHON_39:
            self.assertEqual(feedback[0].message, """Bad indentation on line 1 or adjacent line.

The traceback was:
Line 1 of file answer.py
    def x():
""" + ('            ^\n' if IS_AT_LEAST_PYTHON_311 else "") + """
""" + ('Expected an indented block after function definition on line 1.' if IS_AT_LEAST_PYTHON_310 else "Expected an indented block.") + """
Suggestion: Check line 1, the line before it, and the line after it. Ignore blank lines.""")
        else:
            self.assertEqual(feedback[0].message, """Bad syntax on line 1.

The traceback was:
Line 1 of file answer.py
    def x():

Unexpected eof while parsing.
Suggestion: Check line 1, the line before it, and the line after it. Ignore blank lines.""")

    def test_sections_syntax_errors(self):
        contextualize_report(dedent('''
        NAMES = [____, ____]
        ##### Part 1
        a = 0
        print(a)
        ##### Part 2
        Syntax Error!
        ##### Part 3
        Runtime Error
        '''))
        separate_into_sections(independent=True)
        check_section_exists(3)
        self.assertEqual(get_all_feedback()[0].label, "FeedbackSourceSection")
        next_section()
        verify_section()
        self.assertEqual(get_all_feedback()[1].label, "FeedbackSourceSection")
        next_section()
        verify_section()
        feedback = get_all_feedback()
        self.assertTrue(feedback)
        self.assertEqual(feedback[3].label, "syntax_error")
        print(feedback[0].location)
        self.assertEqual(feedback[3].location.line, 7)

    def test_damaged_sections(self):
        contextualize_report(dedent('''
        NAMES = [____, ____]
        ##### Part 1
        a = 0
        print(a)
        ##### Part 2
        Syntax Error!
        #### Part 3
        Runtime Error
        '''))
        verify()
        separate_into_sections()
        check_section_exists(3)
        feedback = get_all_feedback()
        self.assertTrue(feedback)
        self.assertTrue(feedback[0].label, "incorrect_number_of_sections")

    def test_sections_tifa(self):
        contextualize_report(dedent('''
        ##### Part 1
        a = 0
        ##### Part 2
        print(a)
        ##### Part 3
        print(b)
        '''))
        separate_into_sections(independent=False)
        # First section has an unused variable
        next_section()
        self.assertEqual(len(get_all_feedback()), 2)
        tifa_analysis()
        self.assertEqual(len(get_all_feedback()), 3)
        # Second section uses said variable
        next_section()
        self.assertEqual(len(get_all_feedback()), 4)
        tifa_analysis()
        self.assertEqual(len(get_all_feedback()), 4)
        # Third section has a new unused variables
        next_section()
        self.assertEqual(len(get_all_feedback()), 5)
        tifa_analysis()
        feedback = get_all_feedback()
        self.assertEqual(len(get_all_feedback()), 6)
        finals = sectional.resolve()
        self.assertEqual("""FeedbackSourceSection
Feedback separated into groups
Unused Variable
The variable a was given a value on line 3, but was never used after that.
Initialization Problem
The variable b was used on line 7, but it was not given a value on a previous line. You cannot use a variable until it has been given a value.""",
                         "\n".join(f.title+"\n"+f.message for f in finals.values()))


if __name__ == '__main__':
    unittest.main(buffer=False)



import unittest
import os
import sys
from textwrap import dedent
from pprint import pprint

from pedal.source import set_source

from pedal.cait import parse_program
from pedal.sandbox.commands import run

pedal_library = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, pedal_library)

from pedal import *
from pedal.core import *
from pedal.resolvers import sectional
from pedal.questions import *
from pedal.assertions import *

from tests.execution_helper import Execution, ExecutionTestCase


class TestQuestions(ExecutionTestCase):

    def test_choose_ask(self):
        with Execution('0') as e:
            set_seed(0)
            question_A = Question("QA", "Create a for loop.", [lambda q: False])
            question_B = Question("QB", "Create an if statement.", [lambda q: False])
            pool_1 = Pool("P1", [question_A, question_B])
            pool_1.choose().ask()
        self.assertFeedback(e, "Show Question\nCreate a for loop.")
        
        with Execution('0') as e:
            set_seed(1)
            question_A = Question("QA", "Create a for loop.", [lambda q: False])
            question_B = Question("QB", "Create an if statement.", [lambda q: False])
            pool_1 = Pool("P1", [question_A, question_B])
            pool_1.choose().ask()
        self.assertFeedback(e, "Show Question\nCreate an if statement.")

    def test_choose_ask_using_with(self):
        with Execution('0') as e:
            set_seed(0)
            with Pool("P2") as pool_2:
                with Question("QA", "Create a for loop") as question_A:
                    gently("No answer")
                with Question("QB", "Create an if statement") as question_B:
                    gently("Different answer")
            pool_2.ask()
        self.assertFeedback(e, "Instructor Feedback\nNo answer")

    def test_function_grader_tifa(self):
        # Defaults to ignoring the grading function
        with Execution('def x(a: int) -> int:\n return a') as e:
            fg = FunctionGrader("x", [1, [int], int], [ [[1], 1], [[3], 3] ])
            pool_1 = Pool("P1", [Question("Q1", "Identity function", fg)])
            pool_1.choose().ask()
        self.assertFeedback(e, "Complete\nGreat work!")

        # Detects when not suppressed
        with Execution('def x(a: int) -> int:\n return a') as e:
            fg = FunctionGrader("x", [1, [int], int], [ [[1], 1], [[3], 3] ], {
                'suppress_function_unused': False
            })
            pool_1 = Pool("P1", [Question("Q1", "Identity function", fg)])
            pool_1.choose().ask()
        self.assertFeedback(e, "Unused Variable\nThe function x was given a definition on line 1, but was never used after that.")

        # Still detects other suppressed variables
        with Execution('def x(a: int) -> int:\n return 0\ny=0') as e:
            fg = FunctionGrader("x", [1, [int], int], [ [[1], 0], [[3], 0] ])
            pool_1 = Pool("P1", [Question("Q1", "Identity function", fg)])
            pool_1.choose().ask()
        self.assertFeedback(e, "Unused Variable\nThe variable y was given a value on line 3, but was never used after that.")

    # TODO: Finish this test once the functionality is ready!

    @unittest.skip("Question system is incomplete, finish testing!")
    def test_exam_progress(self):
        def test_has_loop(question):
            ast = parse_program()
            if not ast.find_all("For"):
                gently("No for loop yet.")
            else:
                question.answer()
        question_A = Question("QA", "Create a for loop.", [test_has_loop])
        question_B = Question("QB", "Create an if statement.", [])
        pool_1 = Pool("P1", [question_A, question_B])
        def test_variable(question):
            student = run()
            if assert_has_variable(student, "alpha", value=2):
                question.answer()
        question_C = Question("QC", "Create a variable.", [test_variable])
        pool_2 = Pool("P2", [question_C])
        def test_second_variable(question):
            student = run()
            if not assert_has_variable(student, "beta", value=3):
                question.answer()
                set_success()
        question_D = Question("QD", "Create a beta variable.", [test_second_variable])
        pool_3 = Pool("P3", [question_D])
        def make_exam(code):
            """

            Args:
                code:
            """
            clear_report()
            set_seed(0)
            set_source(code)
            student = run()
            pool_1.choose().ask()
            pool_2.choose().ask()
            if pool_1.answered and pool_2.answered:
                pool_3.choose().ask()
        # Opened a blank file and evaluated
        make_exam('')
        finals = sectional.resolve()
        print(finals)
        self.assertEqual(finals[0][0]['message'], "Create a for loop.")
        self.assertEqual(finals[0][1]['message'], "Create a variable.")
        self.assertEqual(finals[0][2]['message'], "Source code file is blank.")
        # Tried writing some inadequate code
        make_exam('if True: pass')
        finals = sectional.resolve()
        self.assertEqual(finals[0][0]['message'], "Create a for loop.")
        # Tried writing some bad code
        make_exam('1 + ""')
        finals = sectional.resolve()
        self.assertEqual(finals[0][0]['message'], "Create a for loop.")
        self.assertEqual(finals[0][1]['message'], "Create a variable.")
        self.assertEqual(finals[0][2]['message'], dedent("""
               A TypeError occurred:

<pre>Unsupported operand type(s) for +: 'int' and 'str'</pre>

I ran the file `answer.py`.

The traceback was:
  Line 1 of file answer.py
    1 + ""


Type errors occur when you use an operator or function on the wrong type of value. For example, using `+` to add to a list (instead of `.append`), or dividing a string by a number.

Suggestion: To fix a type error, you should trace through your code. Make sure each expression has the type you expect it to have.""").lstrip())
        # Tried writing some good code
        make_exam('for x in []: pass')
        final_success, final_score, _, finals = sectional.resolve()
        self.assertEqual(finals[0][0]['message'], "Create a variable.")
        self.assertFalse(final_success)
        # Tried writing some 100% correct code
        make_exam('for x in []: pass\nalpha = 2\nbeta = 3')
        final_success, final_score, _, finals = sectional.resolve()
        self.assertEqual(finals[0][0]['message'], "No errors reported.")
        self.assertTrue(final_success)

if __name__ == '__main__':
    unittest.main(buffer=False)



"""
A package of tools for capturing student code by matching it against patterns.
"""

from pedal.cait.cait_api import (find_match, find_matches, find_asts,
                                 parse_program,
                                 find_submatches, find_expr_sub_matches,
                                 def_use_error, data_state, data_type,
                                 expire_cait_cache)
from pedal.cait.constants import TOOL_NAME



"""
Tests for checking that runtime assertions are working as expected.
"""
from pedal.assertions.feedbacks import assert_group
from pedal.assertions.runtime import *
from pedal.sandbox.commands import call, start_trace, get_sandbox, evaluate, CommandBlock, run
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE
from pedal.utilities.system import IS_AT_LEAST_PYTHON_39
import unittest


class TestAssertions(ExecutionTestCase):

    def test_assert_equal_basic_passes(self):
        with Execution('5') as e:
            assert_equal(5, 5)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_equal_basic_fails(self):
        with Execution('5') as e:
            assert_equal(5, 4)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
    5 != 4""")

    def test_assert_equal_missing_function(self):
        with Execution('def add(a, b): return a+b', run_tifa=False) as e:
            assert_equal(e.student.call('minus', 1, 2), 3)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
The following exception occurred:
The function minus does not exist.""")

    def test_assert_equal_call_left_fails(self):
        with Execution('def add(a, b): return a-b', run_tifa=False) as e:
            assert_equal(e.student.call('add', 1, 2), 3)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    add(1, 2)
The value of the result was:
    -1
But I expected the result to be equal to:
    3""")

    def test_assert_equal_call_left_passes(self):
        with Execution('def add(a, b): return a+b', run_tifa=False) as e:
            assert_equal(e.student.call('add', 1, 2), 3)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_equal_call_right_fails(self):
        with Execution('def add(a, b): return a-b', run_tifa=False) as e:
            assert_equal(3, e.student.call('add', 1, 2))
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    add(1, 2)
The value of the result was:
    -1
But I expected the result to be equal to:
    3""")

    def test_assert_equal_call_right_passes(self):
        with Execution('def add(a, b): return a+b', run_tifa=False) as e:
            assert_equal(3, e.student.call('add', 1, 2))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_in_call_left_fails(self):
        with Execution('def make_int(): return 7', run_tifa=False) as e:
            assert_in(e.student.call('make_int'), [1, 2, 3])
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    make_int()
The value of the result was:
    7
But I expected the result to be in:
    [1, 2, 3]""")

    def test_assert_in_call_right_fails(self):
        with Execution('def make_ints(): return [1,2,3]', run_tifa=False) as e:
            assert_in(10, e.student.call('make_ints'))
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    make_ints()
The value of the result was:
    [1, 2, 3]
But I expected the result to contain:
    10""")

    def test_assert_is_none_call_left_fails(self):
        with Execution('def do_math(): return 1+2', run_tifa=False) as e:
            assert_is_none(e.student.call('do_math'))
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    do_math()
The value of the result was:
    3
But I expected the result to be None""")

    def test_assert_is_none_call_left_passes(self):
        with Execution('def do_math(): 1+2', run_tifa=False) as e:
            assert_is_none(e.student.call('do_math'))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_is_call_left_passes(self):
        with Execution('r = [1, 2]\ndef get_list(): return r', run_tifa=False) as e:
            result_list = e.student.call('get_list')
            value_list = e.student.data['r']
            assert_is(result_list, value_list)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_is_basic_passes(self):
        with Execution('x = 0\nx is x', run_tifa=False) as e:
            l = [1,2,3]
            r = l
            assert_is(l, r)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_is_call_left_fails(self):
        with Execution('r = [1, 2]\ndef get_list(): return r', run_tifa=False) as e:
            result_list = e.student.call('get_list')
            assert_is(result_list, [1, 2])
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    get_list()
The value of the result was:
    [1, 2]
But I expected the result to be identical to:
    [1, 2]""")

    def test_assert_length_equal_basic_passes(self):
        with Execution('5') as e:
            assert_length_equal([1,2,3], 3)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_length_equal_basic_fails(self):
        with Execution('5') as e:
            assert_length_equal([1, 2, 3], 4)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
    [1, 2, 3] did not have the length 4""")

    def test_assert_length_equal_call_left_fails(self):
        with Execution('def get(a, b): return [a,b]', run_tifa=False) as e:
            assert_length_equal(e.student.call('get', 1, 2), 3)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    get(1, 2)
The value of the result was:
    [1, 2]
But I expected the result to have the length:
    3""")

    def test_assert_length_equal_call_left_passes(self):
        with Execution('def get(a, b): return "test"', run_tifa=False) as e:
            assert_length_equal(e.student.call('get', 1, 2), 4)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_length_equal_call_right_fails(self):
        with Execution('def get(): return 5', run_tifa=False) as e:
            assert_length_equal([1,2,3], e.student.call('get'))
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    get()
The value of the result was:
    5
But I expected the result to be the length of:
    [1, 2, 3]""")

    def test_assert_length_equal_call_right_passes(self):
        with Execution('def get(): return 3', run_tifa=False) as e:
            assert_length_equal([1, 2, 3], e.student.call('get'))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_contains_subset_call_left_passes(self):
        with Execution('def get(): return [1,2,3]', run_tifa=False) as e:
            assert_contains_subset(e.student.call('get'), [1, 2, 3, 4, 5])
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_contains_subset_call_left_fails(self):
        with Execution('def get(): return [1,2,3]', run_tifa=False) as e:
            assert_contains_subset(e.student.call('get'), [1,2])
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    get()
The value of the result was:
    [1, 2, 3]
But I expected the result to be in:
    [1, 2]""")

    def test_assert_not_contains_subset_call_left_passes(self):
        with Execution('def get(): return [1,2,3,4,5]', run_tifa=False) as e:
            assert_not_contains_subset(e.student.call('get'), [1, 2, 3])
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_not_contains_subset_call_left_fails(self):
        with Execution('def get(): return [1,2]', run_tifa=False) as e:
            assert_not_contains_subset(e.student.call('get'), [1, 2, 3])
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    get()
The value of the result was:
    [1, 2]
But I expected the result to not be in:
    [1, 2, 3]""")

    def test_assert_output_call_left_passes(self):
        with Execution('def hi(): print("Hello world!")', run_tifa=False) as e:
            assert_output(e.student.call('hi'), "Hello world!")
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_output_call_left_fails(self):
        with Execution('def hi(): print("Hello world!")', run_tifa=False) as e:
            assert_output(e.student.call('hi'), "Oh Hi There")
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    hi()
The function printed:
    Hello world!
But I expected the output to be:
    Oh Hi There""")

    def test_assert_has_function_passes(self):
        with Execution('def hi(): print("Hello world!")', run_tifa=False) as e:
            assert_has_function(e.student, "hi")
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_data_variable_fails(self):
        with Execution('alpha = 4', run_tifa=False) as e:
            assert_equal(e.student['alpha'], 5)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran your code.
The value of alpha was:
    4
But I expected alpha to be equal to:
    5""")

    def test_assert_group_fails_some_errors(self):
        with Execution('def add(a, b): return a+b', run_tifa=False) as e:
            with assert_group('add') as g:
                assert_equal(e.student.call('add', 1, 2), 3)
                assert_equal(e.student.call('add', 1, 4), 6)
                assert_equal(e.student.call('add', 1, "2"), 3)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor tests.
You passed 1/3 tests.

I ran your function add on some new arguments.
  |  Arguments |                                           Returned | Expected
==============================================================================
  |       1, 2 |                                                  3 |        3
× |       1, 4 |                                                  5 |        6
× |     1, '2' | unsupported operand type(s) for +: 'int' and 'str' |        3""")

    def test_assert_group_fails_all(self):
        with Execution('def add(a, b): return a+b', run_tifa=False) as e:
            with assert_group('add') as g:
                assert_equal(e.student.call('add', 1, 3), 3)
                assert_equal(e.student.call('add', 1, 4), 6)
                assert_equal(e.student.call('add', 1, 3), 3)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor tests.
You passed 0/3 tests.

I ran your function add on some new arguments.
  | Arguments | Returned | Expected
===================================
× |      1, 3 |        4 |        3
× |      1, 4 |        5 |        6
× |      1, 3 |        4 |        3""")

    def test_assert_group_passes(self):
        with Execution('def add(a, b): return a+b', run_tifa=False) as e:
            with assert_group('add') as g:
                assert_equal(e.student.call('add', 1, 3), 4)
                assert_equal(e.student.call('add', 1, 4), 5)
                assert_equal(e.student.call('add', 1, 3), 4)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_coverage_fails(self):
        with Execution('def add(a, b):\n  return a+b', run_tifa=False, tracer_style='native') as e:
            ensure_coverage(.9)
        self.assertFeedback(e, """You Must Test Your Code
Your code coverage is not adequate. You must cover at least 90% of your code to receive feedback. So far, you have only covered 50%.""")

    def test_assert_coverage_passes(self):
        with Execution('def add(a, b):\n  return a+b\nadd(1,2)', run_tifa=False, tracer_style='native') as e:
            ensure_coverage()
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_group_passes_but_earlier_fails(self):
        with Execution('def add(a, b): return a+b', run_tifa=False) as e:
            assert_equal(e.student.call('add', 4, 1), 7)
            with assert_group('add') as g:
                assert_equal(e.student.call('add', 1, 3), 4)
                assert_equal(e.student.call('add', 1, 4), 5)
                assert_equal(e.student.call('add', 1, 3), 4)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    add(4, 1)
The value of the result was:
    5
But I expected the result to be equal to:
    7""")

    def test_assert_equal_variable_passes(self):
        with Execution('x=7\nprint(x)') as e:
            assert_has_variable(e.student, 'x')
            assert_equal(e.student.data['x'], 7)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_equal_variable_fails_does_not_exist(self):
        with Execution('y=7\nprint(y)') as e:
            assert_has_variable(e.student, 'x')
            assert_equal(e.student.data.get('x'), 7)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran your code.
The variable x was not created.""")

    def test_assert_equal_variable_fails_wrong_value(self):
        with Execution('x=3\nprint(x)') as e:
            assert_has_variable(e.student, 'x')
            assert_equal(evaluate('x'), 7)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I evaluated the expression:
    x
The value of the result was:
    3
But I expected the result to be equal to:
    7""")


    def test_assert_type_passes_int(self):
        with Execution('x=3\nprint(x)') as e:
            assert_type(evaluate('x'), int)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_type_fails_int(self):
        with Execution('x="3"\nprint(x)') as e:
            assert_type(evaluate('x'), int)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I evaluated the expression:
    x
The value of the result was:
    'a string'
But I expected the result to be a value of type:
    'an integer'""")

    @unittest.skipUnless(IS_AT_LEAST_PYTHON_39, "requires python 3.9")
    def test_assert_type_fails_non_generic_list(self):
        with Execution('x=["3"]\nprint(x)') as e:
            assert_type(evaluate('x'), list[int])
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I evaluated the expression:
    x
The value of the result was:
    'a list of strings'
But I expected the result to be a value of type:
    'a list of integers'""")

    def test_assert_type_passes_instructor_dataclass(self):
            with Execution('''from dataclasses import dataclass
@dataclass
class Dog:
    name: str
    age: int
x = Dog("D", 5)
print(x)''') as e:
                from dataclasses import dataclass
                @dataclass
                class Dog:
                    name: str
                    age: int
                assert_type(evaluate('x'), Dog)
            self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_type_passes_instructor_dataclass(self):
        with Execution('''from dataclasses import dataclass
@dataclass
class Dog:
    name: str
    age: int
x = Dog("D", 5)
print(x)''') as e:
            from dataclasses import dataclass
            @dataclass
            class Dog:
                pass

            assert_type(evaluate('x'), Dog)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_type_fails_instructor_dataclass(self):
        with Execution('''from dataclasses import dataclass
@dataclass
class Dog:
    name: str
    age: int
x = Dog("D", 5)
print(x)''') as e:
            from dataclasses import dataclass
            @dataclass
            class Cat:
                pass

            assert_type(evaluate('x'), Cat)
        self.assertFeedback(e, """Name Error
I evaluated the expression:
    Cat

A NameError occurred:

    Name 'Cat' is not defined


A name error means you have used a variable that has no value.  You may have a typo, so check the spelling. Or, you may have forgotten to initialize a variable.

Suggestion: Trace your code and make sure each variable was set before it was read.""")

    @unittest.skipUnless(IS_AT_LEAST_PYTHON_39, "requires python 3.9+")
    def test_assert_type_passes_instructor_dataclass_nested(self):
        with Execution('''from dataclasses import dataclass
@dataclass
class Dog:
    name: str
    age: int
    hobbies: list[str]
x = Dog("D", 5, ["fetch"])
print(x)''') as e:
            from dataclasses import dataclass
            @dataclass
            class Dog:
                pass

            assert_equal(evaluate('x'), call('Dog', 'D', 5, ['fetch']))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_chain_passes(self):
        with Execution('x=3\nprint(x)') as e:
            assert_has_variable(e.student, 'x')
            assert_type(evaluate('x'), int)
            assert_equal(evaluate('x'), 3)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_chain_missing_variable(self):
        with Execution('y=3\nprint(y)') as e:
            assert_has_variable(e.student, 'x')
            assert_type(evaluate('x'), int)
            assert_equal(evaluate('x'), 3)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran your code.
The variable x was not created.""")

    def test_assert_chain_incompatible_type(self):
        with Execution('x=3\nprint(x)') as e:
            assert_has_variable(e.student, 'x')
            assert_type(evaluate('x'), list)
            assert_equal(evaluate('set(x)'), {})
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I evaluated the expression:
    x
The value of the result was:
    'an integer'
But I expected the result to be a value of type:
    'a list'""")

    def test_command_block(self):
        with Execution('''
class Fruit:
    def __init__(self, name, weight=0):
        self.name = name
        self.weight = weight
def do_math(a, b):
    return a + b - 5
def weigh_fruits(fruits):
    return sum(fruit.weight for fruit in fruits)    
                ''', run_tifa=False) as e:

            with CommandBlock():
                orange = call("Fruit", "Orange", 30, target="orange")
                self.assertIsInstance(orange, e.student.data['Fruit'])
                pineapple = call("Fruit", "Pineapple", 60, target="pineapple")
                run("fruits = [orange, pineapple]")
                total_weight = call('weigh_fruits', args_locals=["fruits"])
                assert_equal(evaluate('pineapple.weight'), 61)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    orange = Fruit('Orange', 30)
    pineapple = Fruit('Pineapple', 60)
    fruits = [orange, pineapple]
    weigh_fruits(fruits)
I evaluated the expression:
    pineapple.weight
The value of the result was:
    60
But I expected the result to be equal to:
    61""")

    def test_empty_string_input(self):
        with Execution('''
def adder():
    return input() + input()
adder
''') as e:
            assert_equal(call('adder', inputs=['A', '', 'B']), 'AB')
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    adder()

And I entered as input:
    A
     
The value of the result was:
    'A'
But I expected the result to be equal to:
    'AB'""")

    def test_assert_regex_passes(self):
        with Execution('def message(): return "Hello World!"\nmessage') as e:
            assert_regex("World!?", call('message'))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_regex_fails(self):
        with Execution('def message(): return "Banana"\nmessage') as e:
            assert_regex("World!?", call('message'))
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    message()
The value of the result was:
    'Banana'
But I expected the result to match the text:
    'World!?'""")

    def test_assert_not_regex_fails(self):
        with Execution('def message(): return "Hello World!"\nmessage') as e:
            assert_not_regex("World!?", call('message'))
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran the code:
    message()
The value of the result was:
    'Hello World!'
But I expected the result to not match the text:
    'World!?'""")

    def test_assert_not_regex_passes(self):
        with Execution('def message(): return "Banana"\nmessage') as e:
            assert_not_regex("World!?", call('message'))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_output_regex_passes(self):
        with Execution('print("Hello World!")') as e:
            assert_output_regex("World!?", e.student)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_output_regex_fails(self):
        with Execution('print("Banana")') as e:
            assert_output_regex("World!?", e.student)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran your code.
The printed output was:
    Banana
But I expected the output to match the regex:
    World!?""")

    def test_assert_not_output_regex_fails(self):
        with Execution('print("Hello World!")') as e:
            assert_not_output_regex("World!?", e.student)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor test.
I ran your code.
The printed output was:
    Hello World!
But I expected the output to not match the regex:
    World!?""")

    def test_assert_not_output_regex_passes(self):
        with Execution('print("Banana")') as e:
            assert_not_output_regex("World!?", e.student)
        self.assertFeedback(e, SUCCESS_MESSAGE)


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Assertions for working with various testing libraries. Currently we support:
- bakery (formerly known as cisc106.py and cisc108.py)

We do intend to add more support as people request. If you have a testing
library that you would like to see supported, please open an issue on GitHub!

# TODO: At least support the builtin unittest library.
"""
from pedal import MAIN_REPORT
from pedal.core.feedback import CompositeFeedbackFunction
from pedal.core.commands import gently
from pedal.sandbox.commands import get_student_data, run, call, evaluate

# TODO: Decompose these with a helper function that parameterizes these fields.


@CompositeFeedbackFunction()
def ensure_cisc108_tests(test_count, **kwargs):
    """
    Ensure that the student has not failed their own tests.
    This is for the specific ``cisc108`` library, not the general unittest
    library.
    """
    student = get_student_data()
    if 'assert_equal' not in student:
        return gently(f"You have not imported assert_equal from the cisc108 module.",
               label=f"missing_cisc108_assert_equal",
               title="Missing assert_equal", **kwargs)
    assert_equal = student['assert_equal']
    if not hasattr(assert_equal, 'student_tests'):
        return gently("The assert_equal function has been modified. Do not let it be overwritten!",
               label="overwrote_assert_equal",
               title="Overwrote assert_equal", **kwargs)
    student_tests = assert_equal.student_tests
    if student_tests.tests == 0:
        return gently("You are not unit testing the result.",
                      title="No Student Unit Tests",
                      label="no_student_tests", **kwargs)
    elif student_tests.tests < test_count:
        return gently("You have not written enough unit tests.",
                      label="not_enough_tests",
                      title="Not Enough Student Unit Tests", **kwargs)
    elif student_tests.failures > 0:
        failures = student_tests.failures
        successes = student_tests.successes
        tests = student_tests.tests
        return gently(f"{failures}/{tests} of your unit tests are not passing.",
                      label="failing_student_tests",
                      title="Student Unit Tests Failing",
                      fields={'failures': failures, 'successes': successes,
                              'tests': tests},
                      **kwargs)
    return False


@CompositeFeedbackFunction()
def ensure_cisc106_tests(test_count, **kwargs):
    """
    Ensure that the student has not failed their own tests.
    This is for the specific ``cisc106`` library, not the general unittest
    library.
    """
    student = get_student_data()
    if 'assertEqual' not in student:
        return gently("You have not imported assertEqual from the cisc106 module.",
               label="missing_cisc106_assertEqual",
               title="Missing assertEqual", **kwargs)
    assertEqual = student['assertEqual']
    if not hasattr(assertEqual, 'student_tests'):
        return gently("The assertEqual function has been modified. Do not let it be overwritten!",
               title="overwrote_assertEqual",
               label="Overwrote assertEqual", **kwargs)
    student_tests = assertEqual.student_tests
    if student_tests.tests == 0:
        return gently("You are not unit testing the result.",
                      title="No Student Unit Tests",
                      label="no_student_tests", **kwargs)
    elif student_tests.tests < test_count:
        return gently("You have not written enough unit tests.",
                      label="not_enough_tests",
                      title="Not Enough Student Unit Tests", **kwargs)
    elif student_tests.failures > 0:
        failures = student_tests.failures
        successes = student_tests.successes
        tests = student_tests.tests
        return gently(f"{failures}/{tests} of your unit tests are not passing.",
                      label="failing_student_tests",
                      title="Student Unit Tests Failing",
                      fields={'failures': failures, 'successes': successes,
                              'tests': tests},
                      **kwargs)
    return False


@CompositeFeedbackFunction()
def ensure_bakery_tests(test_count, **kwargs):
    """
    Ensure that the student has not failed their own tests.
    This is for the specific ``bakery`` library, not the general unittest
    library.
    """
    student = get_student_data()
    if 'assert_equal' not in student:
        return gently(f"You have not imported assert_equal from the `bakery` module.",
               label=f"missing_bakery_assert_equal",
               title="Missing assert_equal", **kwargs)
    assert_equal = student['assert_equal']
    if not hasattr(assert_equal, 'student_tests'):
        return gently("The assert_equal function has been modified. Do not let it be overwritten!",
               label="overwrote_assert_equal",
               title="Overwrote assert_equal", **kwargs)
    student_tests = assert_equal.student_tests
    if student_tests.tests == 0:
        return gently("You are not unit testing the result.",
                      title="No Student Unit Tests",
                      label="no_student_tests", **kwargs)
    elif student_tests.tests < test_count:
        return gently("You have not written enough unit tests.",
                      label="not_enough_tests",
                      title="Not Enough Student Unit Tests", **kwargs)
    elif student_tests.failures > 0:
        failures = student_tests.failures
        successes = student_tests.successes
        tests = student_tests.tests
        return gently(f"{failures}/{tests} of your unit tests are not passing.",
                      label="failing_student_tests",
                      title="Student Unit Tests Failing",
                      fields={'failures': failures, 'successes': successes,
                              'tests': tests},
                      **kwargs)
    return False


@CompositeFeedbackFunction()
def ensure_unittest_tests(test_count=None, report=MAIN_REPORT, **kwargs):
    """
    Ensure that the student has not failed their own tests.
    This is for the specific ``unittest`` library.

    Args:
        test_count (int): The number of tests that the student should have written. If `None`, any number is fine.
    """
    import unittest
    # Call the unittest.main just in case it wasn't explicitly run
    # Check whether the tests were successful
    # Check that the right number of tests were run
    # TODO: Make all of this more flexible, if people request it.

    student = get_student_data()
    if 'unittest' not in student:
        run("import unittest", report=report)

    test_names = {name for name, value in student.items()
                  if isinstance(value, type) and issubclass(value, unittest.TestCase)}
    result = run(f"""
def __load_tests():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    for test_class in [{", ".join(test_names)}]:
        tests = loader.loadTestsFromTestCase(test_class)
        suite.addTests(tests)
    return suite
__test_result = unittest.TestResult()
__test_suite = __load_tests()
__test_suite.run(__test_result)
""", report=report)
    student_tests: unittest.TestResult = evaluate("__test_result",
                                                  report=report)
    if test_count and student_tests.testsRun == 0:
        return gently("You are not unit testing the result.",
                      title="No Student Unit Tests",
                      label="no_student_tests", **kwargs)
    elif test_count is not None and student_tests.testsRun < test_count:
        return gently("You have not written enough unit tests.",
                      label="not_enough_tests",
                      title="Not Enough Student Unit Tests", **kwargs)
    elif len(student_tests.failures) + len(student_tests.errors) > 0:
        failures = len(student_tests.failures)
        errors = len(student_tests.errors)
        tests = student_tests.testsRun
        successes = tests - (failures + errors)
        return gently(f"{failures+errors}/{tests} of your unit tests are not passing.",
                      label="failing_student_tests",
                      title="Student Unit Tests Failing",
                      fields={'failures': failures + errors, 'successes': successes,
                              'tests': tests},
                      **kwargs)
    return False

@CompositeFeedbackFunction()
def ensure_pytest_tests(test_count=None, report=MAIN_REPORT, **kwargs):
    """
    Ensure that the student has not failed their own tests.
    This is for the specific ``pytest`` library.

    Args:
        test_count (int): The number of tests that the student should have written. If `None`, any number is fine.
    """
    # Call the unittest.main just in case it wasn't explicitly run
    # Check whether the tests were successful
    # Check that the right number of tests were run
    # TODO: Make all of this more flexible, if people request it.

    student = get_student_data()

    test_names = {name for name, value in student.items() if name.startswith("test")}
    result = run(f"""
__successes, __failures = 0, 0
for __a_test in [{", ".join(test_names)}]:
    try:
        __a_test()
        __successes += 1
    except:
        __failures += 1
""", report=report)
    successes = evaluate("__successes", report=report)
    failures = evaluate("__failures", report=report)
    tests_run = successes+failures
    if test_count and tests_run == 0:
        return gently("You are not unit testing the result.",
                      title="No Student Unit Tests",
                      label="no_student_tests", **kwargs)
    elif test_count is not None and tests_run < test_count:
        return gently("You have not written enough unit tests.",
                      label="not_enough_tests",
                      title="Not Enough Student Unit Tests", **kwargs)
    elif failures + successes > 0:
        return gently(f"{failures}/{tests_run} of your unit tests are not passing.",
                      label="failing_student_tests",
                      title="Student Unit Tests Failing",
                      fields={'failures': failures, 'successes': successes,
                              'tests': tests_run},
                      **kwargs)
    return False


from pedal.types.new_types import ModuleType, void_function, register_builtin_module


def load_turtle_module():
    return ModuleType('turtle', {
        'forward': void_function('forward'),
        'backward': void_function('backward'),
        'color': void_function('color'),
        'right': void_function('right'),
        'left': void_function('left'),
    }, {})


register_builtin_module('turtle', load_turtle_module)



"""
Collection of feedback functions related to static checks of student code.
"""
try:
    import dataclasses
except ImportError:
    dataclasses = None
import ast

from pedal.assertions.classes import (missing_dataclass, dataclass_not_available,
                                      missing_field_type, too_few_fields, too_many_fields,
                                      duplicate_dataclass_definition, invalid_field_type,
                                      wrong_fields_type, name_is_not_a_dataclass, missing_dataclass_annotation,
                                      unknown_field, wrong_field_order)
from pedal.assertions.functions import *
from pedal.cait.cait_api import parse_program, find_match, find_matches
from pedal.assertions.feedbacks import AssertionFeedback
from pedal.cait.find_node import find_operation, find_function_calls, find_function_definition
from pedal.core.feedback import CompositeFeedbackFunction, Feedback, FeedbackResponse
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.core.commands import compliment as core_compliment, give_partial, explain, gently
from pedal.tifa.commands import tifa_type_check
from pedal.types.normalize import normalize_type
from pedal.types.new_types import is_subtype
from pedal.utilities.ast_tools import AST_NODE_NAMES


class EnsureAssertionFeedback(AssertionFeedback):
    """ Abstract base class for assertions preventing things. """
    def __init__(self, name, at_least=1, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        root_message = "" if not root else " (inside of some other code)"
        fields = {'name': name, 'at_least': at_least, 'capacity': '',
                  'root': root, 'name_message': report.format.name(name),
                  'root_message': root_message}
        super().__init__(fields=fields, **kwargs)

    def _check_usage(self, field_name, uses):
        at_least = self.fields['at_least']
        self.fields[field_name] = use_count = len(uses)
        if at_least > use_count:
            if at_least == 1:
                return True
            else:
                self.fields['capacity'] = (f" at least {at_least} times, but"
                                           f" you used it {use_count} times")
                return True
        return False


class PreventAssertionFeedback(AssertionFeedback):
    """ Abstract base class for assertions preventing things. """
    def __init__(self, name, at_most=0, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        root_message = "" if not root else " (inside of some other code)"
        fields = {'name': name, 'at_most': at_most, 'capacity': '',
                  'root': root, 'name_message': report.format.name(name),
                  'root_message': root_message}
        super().__init__(fields=fields, **kwargs)

    def _check_usage(self, field_name, uses):
        at_most = self.fields['at_most']
        self.fields[field_name] = use_count = len(uses)
        if use_count and at_most < use_count:
            if at_most == 0:
                return True
            else:
                self.fields['capacity'] = (f" more than {at_most} times, but"
                                           f" you used it {use_count} times")
                return True
        return False


class prevent_function_call(PreventAssertionFeedback):
    """
    Determine if a function was ever called by statically checking
    the students' code.

    # TODO: also check the TIFA records to see if the function was read

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to call this
            function. Defaults to ``0``.
    """

    title = "May Not Use Function"
    message_template = ("You used the function {name_message} on line "
                        "{location.line}. You may not use that function"
                        "{capacity}.")

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        calls = find_function_calls(name, root=self.fields['root'])
        if calls:
            self.update_location(Location.from_ast(calls[-1]))
        return self._check_usage('call_count', calls)


class ensure_function_call(EnsureAssertionFeedback):
    """
    Determine that a function was definitely called by statically checking
    the students' code.

    # TODO: Allow parameter to force this to be a method xor/or function call
    # TODO: Allow set to be passed in for OR, or list for AND, for multiples

    # TODO: also check the TIFA records to see if the function was read

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to call this
            function. Defaults to ``0``.
    """

    title = "Must Use Function"
    message_template = "You must use the function {name_message}{capacity}."

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        calls = find_function_calls(name, root=self.fields['root'])
        return self._check_usage('call_count', calls)


class prevent_operation(PreventAssertionFeedback):
    """
    Determines if the given operator `op_name` is not used anywhere.
    Otherwise, returns `None`. You can specify the
    operator as a string like `"+"` or `"<<"`. Supports all comparison,
    boolean, binary, and unary operators.

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to call this
            function. Defaults to ``0``.
    """
    title = "May Not Use Operator"
    message_template = ("You used the operator {name_message} on line "
                        "{location.line}. You may not use that operator"
                        "{capacity}.")

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = self.report.format.python_expression(name)
        uses = list(find_operation(name, root=self.fields['root']))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


class ensure_operation(EnsureAssertionFeedback):
    """
    Determines if the given operator `op_name` is used somewhere. You can
    specify the operator as a string like `"+"` or `"<<"`. Supports all
    comparison, boolean, binary, and unary operators.

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_least (int): The minimum number of times you must call this
            function. Defaults to ``1``.
    """
    title = "Must Use Operator"
    message_template = "You must use the operator {name_message}{capacity}."

    def condition(self):
        """ Use find_matches to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = self.report.format.python_expression(name)
        uses = list(find_operation(name, root=self.fields['root']))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


ensure_operator = ensure_operation
prevent_operator = prevent_operation


class prevent_literal(PreventAssertionFeedback):
    """ Make sure that the given literal value does not appear in the student's
    code. """
    title = "May Not Use Literal Value"
    message_template = ("You used the literal value {literal_message:python_expression} on line "
                        "{location.line}. You may not use that value"
                        "{capacity}.")

    def __init__(self, literal, at_most=0, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal': literal, 'at_most': at_most, 'capacity': '',
                  'root': root, 'literal_message': repr(literal)}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal = self.fields['literal']
        uses = self.fields['root'].find_matches(repr(literal))
        if uses:
            self.update_location(uses[-1].match_location(False))
        return self._check_usage('use_count', uses)


# TODO: ensure_literal does not recognize None?
class ensure_literal(EnsureAssertionFeedback):
    """ Make sure that the given literal value does appear in the student's
        code. """
    title = "Must Use Literal Value"
    message_template = "You must use the literal value {literal_message}{capacity}."

    def __init__(self, literal, at_least=1, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal': literal, 'at_least': at_least, 'capacity': '',
                  'root': root, 'literal_message': report.format.python_expression(repr(literal))}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal = self.fields['literal']
        uses = self.fields['root'].find_matches(repr(literal))
        if uses:
            self.update_location(uses[-1].match_location(False))
        return self._check_usage('use_count', uses)


class prevent_literal_type(PreventAssertionFeedback):
    """ Make sure that the given literal value does not appear in the student's
    code. """
    title = "May Not Use Type of Literal Value"
    message_template = ("You used the literal value type "
                        "{literal_type_message:python_expression} on line {location.line}."
                        " You may not use that type of value{capacity}.")

    def __init__(self, literal_type, at_most=0, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal_type': literal_type, 'at_most': at_most,
                  'capacity': '', 'root': root,
                  'literal_type_message': literal_type.__name__}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal_type = self.fields['literal_type']
        if literal_type == bool:
            uses = (self.fields['root'].find_matches("False") +
                    self.fields['root'].find_matches("True"))
            uses = [match.match_root for match in uses]
        elif literal_type == str:
            uses = self.fields['root'].find_all("Str")
        elif literal_type in (int, float):
            uses = [node for node in self.fields['root'].find_all("Num")
                    if isinstance(node.value, literal_type)]
        elif literal_type == list:
            uses = self.fields['root'].find_all("List")
        elif literal_type == dict:
            uses = self.fields['root'].find_all("Dict")
        else:
            raise ValueError(f"Unknown literal type: {literal_type}")
        if uses:
            self.update_location(uses[-1].lineno)
        return self._check_usage('use_count', uses)


class ensure_literal_type(EnsureAssertionFeedback):
    """ Make sure that the given type of literal value does appear in the
    student's code. """
    title = "Must Use Type of Literal Value"
    message_template = "You must use a literal value of type {literal_type_message:python_expression}{capacity}."

    def __init__(self, literal_type, at_least=1, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'literal_type': literal_type, 'at_least': at_least,
                  'capacity': '', 'root': root,
                  'literal_type_message': literal_type.__name__}
        super(AssertionFeedback, self).__init__(fields=fields, **kwargs)

    def condition(self):
        literal_type = self.fields['literal_type']
        if literal_type == bool:
            uses = (self.fields['root'].find_matches("False")+
                    self.fields['root'].find_matches("True"))
            uses = [match.match_root for match in uses]
        elif literal_type == str:
            uses = self.fields['root'].find_all("Str")
        elif literal_type in (int, float):
            uses = [node for node in self.fields['root'].find_all("Num")
                    if isinstance(node.value, literal_type)]
        elif literal_type == list:
            uses = self.fields['root'].find_all("List")
        elif literal_type == dict:
            uses = self.fields['root'].find_all("Dict")
        else:
            raise ValueError(f"Unknown literal type: {literal_type}")
        if uses:
            self.update_location(uses[-1].lineno)
        return self._check_usage('use_count', uses)

class prevent_ast(PreventAssertionFeedback):
    """
    Determines if the given ast `name` is not used anywhere, returning
    the node of it if it is. You can refer to Green Tree Snakes documentation
    for more information about AST Names:

        https://greentreesnakes.readthedocs.io/en/latest/nodes.html

    Args:
        name (str): The name of the function (e.g., ``'sum'``)
        at_most (int): The maximum number of times you are allowed to use this
            AST. Defaults to ``0``.
    """
    title = "May Not Use Code"
    message_template = ("You used {name_message} on line "
                        "{location.line}. You may not use that"
                        "{capacity}.")
    justification_template = ("Incorrectly found a {name_message}{capacity}{root_message}.",
                              "Correctly found a {name_message}{capacity}{root_message}.")

    def condition(self):
        """ Use find_all to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = AST_NODE_NAMES.get(name, name)
        uses = list(self.fields['root'].find_all(name))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


class ensure_ast(EnsureAssertionFeedback):
    """
    Determines if the given ast `name` is used anywhere, returning
    the node of it if it is. You can refer to Green Tree Snakes documentation
    for more information about AST Names:

        https://greentreesnakes.readthedocs.io/en/latest/nodes.html

    Args:
        name (str): The name of the node.
        at_least (int): The minimum number of times you must use this
            node. Defaults to ``1``.
    """
    title = "Must Use Code"
    message_template = "You must use {name_message}{capacity}."
    justification_template = ("Failed to find a {name_message}{capacity}{root_message}.",
                              "Successfully found a {name_message}{capacity}{root_message}.")

    def condition(self):
        """ Use find_all to check number of occurrences. """
        name = self.fields['name']
        self.fields['name_message'] = AST_NODE_NAMES.get(name, name)
        uses = list(self.fields['root'].find_all(name))
        if uses:
            self.update_location(Location.from_ast(uses[-1]))
        return self._check_usage('use_count', uses)


def function_prints(function_name):
    """
    Determine that the print function is called within this function, by
    tracing its code statically. Does not actually verify that the function
    was ever called in practice.

    Args:
        function_name (str): The name of the function to search.

    Returns:

    """
    return ensure_function_call('print',
                                root=find_function_definition(function_name))


def has_import(ast, name):
    """
    Determine if the given module name is in the code.

    Args:
        ast (:py:class:`pedal.cait.cait_node.CaitNode`): The starting point
            to search.
        name (str): The name of the module to match against.

    Returns:
        bool: Whether or not the module name was matched.
    """
    imports = ast.find_all("Import")
    import_froms = ast.find_all("ImportFrom")
    # What about ``import <name>``
    if imports and any(alias._name == name for i in imports for alias in i.names):
        return True
    # What about ``from <name> import *``
    if import_froms and any(i.module == name for i in import_froms):
        return True
    return False


class ensure_import(EnsureAssertionFeedback):
    """
    Verify that the given ``module`` has been imported, by statically
    checking the import statements in the code. The ``at_least`` parameter
    is ignored.
    """
    title = "Must Import Module"
    message_template = "You must import the module {name_message}."

    def condition(self):
        """ Traverse the AST to find matches. """
        name = self.fields['name']
        ast = self.fields['root']
        return not has_import(ast, name)


class prevent_import(PreventAssertionFeedback):
    """
    Verify that the given ``module`` has not been imported, by statically
    checking the import statements in the code. The ``at_most`` parameter
    is ignored.
    """
    title = "May Not Import Module"
    message_template = "You may not import the module {name_message}."

    def condition(self):
        """ Traverse the AST to find matches. """
        name = self.fields['name']
        ast = self.fields['root']
        return has_import(ast, name)


class ensure_documented_functions(AssertionFeedback):
    """ Checks that all of the functions in the source code are documented. """
    title = "Must Document Functions"
    message_template = ("You must document the following function{plural}:"
                        " {names_message}.")

    def __init__(self, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'root': root}
        super().__init__(fields=fields, **kwargs)

    def condition(self):
        """ Traverse the AST to find matches. """
        defs = self.fields['root'].find_all('FunctionDef')
        names = []
        for a_def in defs:
            # Don't have to document constructors (happens in Class Def)
            if a_def.name == "__init__":
                continue
            if (a_def.body and (a_def.body[0].ast_name != "Expr" or
                                a_def.body[0].value.ast_name not in ("Str", "Constant"))):
                names.append(a_def.name)
        self.fields['names'] = names
        self.fields['names_message'] = ", ".join(self.report.format.name(name)
                                                 for name in names)
        self.fields['plural'] = 's' if names else ''
        return bool(names)

"""
New assertions:

prevent_literal(literal_value: Any)
prevent_function_call(function_name: str) # Whether it was detected statically
prevent_operation(operator: str)
prevent_import(module_name: str)
prevent_ast(ast_name: str)

TODO: prevent_source_text(code: str, missing=False, minimum=2, maximum=3)
TODO: prevent_traced_call(function_name: str)# Whether it was detected dynamically
TODO: prevent_name(variable_name: str)
TODO: prevent_assignment(variable_name: str, type: str, value: Any)
TODO: ensure_function

ensure_prints(count)
    ensure_function_call('print', at_least=count)
    ensure_function_call('print', at_most=count)
function_prints => ensure_function_call('print', root=match_signature('name')

"""


# TODO: wait, is this a positive feedback group waiting to happen?

@CompositeFeedbackFunction(missing_function, duplicate_function_definition,
                           too_few_parameters, too_many_parameters,
                           missing_parameter_type, invalid_parameter_type,
                           wrong_parameter_type, wrong_return_type,
                           missing_return_type)
def ensure_function(name, arity=None, parameters=None,
                    returns=None, root=None, compliment=False, **kwargs):
    """ Checks that the function exists and has the right signature. """
    report = kwargs.get('report', MAIN_REPORT)
    root = root or parse_program(report=report)
    defs = root.find_all('FunctionDef')
    defs = [a_def for a_def in defs if a_def._name == name]
    if not defs:
        return missing_function(name, **kwargs)
    if len(defs) > 1:
        lines = [Location.from_ast(a_def) for a_def in defs]
        return duplicate_function_definition(name, lines, **kwargs)
    definition = defs[0]
    # Actual logic
    # 1.2.1 'arity' style - simply checks number of parameters
    if arity is not None or parameters is not None:
        expected_arity = arity if arity is not None else len(parameters)
        actual_arity = len(definition.args.args)
        if actual_arity < expected_arity:
            return too_few_parameters(name, actual_arity, expected_arity, **kwargs)
        elif actual_arity > expected_arity:
            return too_many_parameters(name, actual_arity, expected_arity, **kwargs)
    # 1.2.2 'parameters' style - checks each parameter's name and type
    if parameters is not None:
        actual_parameters = definition.args.args
        for expected_parameter, actual_parameter in zip(parameters, actual_parameters):
            expected_parameter_type = normalize_type(expected_parameter, tifa_type_check).as_type()
            actual_parameter_name = (actual_parameter.id if actual_parameter.id is not None
                                     else actual_parameter.arg)
            if actual_parameter.annotation is None:
                return missing_parameter_type(name, actual_parameter_name,
                                              expected_parameter_type,
                                              **kwargs)
            try:
                actual_parameter_type = normalize_type(actual_parameter.annotation.ast_node,
                                                       tifa_type_check).as_type()
            except ValueError as e:
                return invalid_parameter_type(name, actual_parameter_name,
                                              actual_parameter.annotation,
                                              expected_parameter_type,
                                              **kwargs)
            if not is_subtype(actual_parameter_type, expected_parameter_type):
                return wrong_parameter_type(name, actual_parameter_name,
                                            actual_parameter_type,
                                            expected_parameter_type, **kwargs)
    # 1.2.3. 'returns' style - checks the return type explicitly
    if returns is not None:
        expected_returns = normalize_type(returns, tifa_type_check).as_type()
        if definition.returns is None:
            return missing_return_type(name, expected_returns, **kwargs)
        try:
            actual_returns = normalize_type(definition.returns.ast_node,
                                            tifa_type_check).as_type()
        except ValueError as e:
            return invalid_return_type(name, definition.returns,
                                       expected_returns, **kwargs)
        if not is_subtype(actual_returns, expected_returns):
            return wrong_return_type(name, actual_returns, expected_returns,
                                     **kwargs)
    # Alternatively, returns positive FF?
    if compliment:
        if isinstance(compliment, str):
            core_compliment(compliment, label="function_defined", **kwargs)
        elif compliment is True:
            core_compliment(f"Defined {name}", label="function_defined", **kwargs)
    elif kwargs.get("score"):
        give_partial(kwargs.pop("score"), label="function_defined", **kwargs)

    return None


@CompositeFeedbackFunction(missing_dataclass, duplicate_dataclass_definition,
                           too_few_fields, too_many_fields, invalid_field_type,
                           unknown_field,
                           missing_field_type, wrong_fields_type, name_is_not_a_dataclass,
                           dataclass_not_available, missing_dataclass_annotation)
def ensure_dataclass(name, fields=None, root=None, compliment=False, **kwargs):
    """
    Ensures that the actual definition for the given class exists.
    You can provide either a string and fields, or an instructor-defined version

    Arguments:
        name (str or dataclass): Either the name of the expected dataclass, or a
            dataclass object (not an instance).
        fields (dict): A dictionary of field names to types.
        root (CaitNode): The AST to search through; defaults to program root.
        compliment (str or bool): If True, will use the compliment function to
            provide positive feedback. If a string, will use that string as the
            compliment. If False, will not provide positive feedback.
    """
    if isinstance(name, str):
        # Fields will be a dictionary or class
        if fields is None:
            fields = {}
    # Or assume it is a dataclass provided by the instructor
    else:
        fields = {field.name: field.type for field in dataclasses.fields(name)}
        name = name.__name__
    # Inspect the code for the definition
    report = kwargs.get('report', MAIN_REPORT)
    root = root or parse_program(report=report)
    defs = root.find_all('ClassDef')
    defs = [a_def for a_def in defs if a_def._name == name]
    if not defs:
        return missing_dataclass(name, **kwargs)
    if len(defs) > 1:
        lines = [Location.from_ast(a_def) for a_def in defs]
        return duplicate_dataclass_definition(name, lines, **kwargs)
    definition = defs[0]
    definition_fields = [line for line in definition.body
                         if isinstance(line.ast_node, ast.AnnAssign) and line.simple]
    # TODO: Support ast.Assign nodes for default values
    # 1. Make sure import is present
    missing_import = ensure_import('dataclasses')
    if missing_import:
        return missing_import
    for decorator in definition.decorator_list:
        # TODO: Support the @dataclasses.dataclass style too
        if isinstance(decorator.ast_node, ast.Name) and decorator.id == 'dataclass':
            break
    else:
        return missing_dataclass_annotation(name, **kwargs)
    # 2. Confirm the number of fields
    expected_arity = len(fields)
    actual_arity = len(definition_fields)
    if actual_arity < expected_arity:
        return too_few_fields(name, actual_arity, expected_arity, **kwargs)
    elif actual_arity > expected_arity:
        return too_many_fields(name, actual_arity, expected_arity, **kwargs)
    # 3. checks each field's name and type
    for actual_field in definition_fields:
        actual_field_name = actual_field.target.id
        if actual_field_name not in fields:
            return unknown_field(name, actual_field_name, **kwargs)
        expected_field = fields[actual_field_name]
        expected_field_type = normalize_type(expected_field, tifa_type_check).as_type()
        if actual_field.annotation is None:
            return missing_field_type(name, actual_field_name, expected_field_type, **kwargs)
        try:
            actual_field_type = normalize_type(actual_field.annotation.ast_node,
                                               tifa_type_check).as_type()
        except ValueError as e:
            return invalid_field_type(name, actual_field_name, actual_field.annotation,
                                      expected_field_type, **kwargs)
        if not is_subtype(actual_field_type, expected_field_type):
            return wrong_fields_type(name, actual_field_name, actual_field_type,
                                     expected_field_type, **kwargs)
    for actual_field, expected_field in zip(definition_fields, fields.keys()):
        actual_field_name = actual_field.target.id
        if actual_field_name != expected_field:
            return wrong_field_order(name, actual_field_name, expected_field, **kwargs)
    # Alternatively, returns positive FF?
    if compliment:
        if isinstance(compliment, str):
            core_compliment(compliment, label="dataclass_defined", **kwargs)
        elif compliment is True:
            core_compliment(f"Defined {name}", label="dataclass_defined", **kwargs)
    elif kwargs.get("score"):
        give_partial(kwargs.pop("score"), label="dataclass_defined", **kwargs)

    return None

@CompositeFeedbackFunction(prevent_function_call, ensure_function_call)
def ensure_prints_exactly(count, **kwargs):
    """
    DEPRECATED.

    Used to ensure that the student's code calls the print function a certain
    number of times.

    Args:
        count (int): The minimum and maximum number of times you can call print.

    Returns:
        bool: Whether or not both the maximum and minimum was met.
    """
    return (ensure_function_call('print', at_least=count, **kwargs) and
            prevent_function_call('print', at_most=count, **kwargs))


class ensure_starting_code(AssertionFeedback):
    """Detects if the given code is missing."""
    title = "Don't Change Starting Code"
    message_template = "You have changed or removed the starting code. "

    def __init__(self, code, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'code': code, 'root': root}
        super().__init__(fields=fields, **kwargs)

    def condition(self):
        """ Traverse the AST to find matches. """
        code = self.fields['code']
        ast = self.fields['root']
        return not ast.find_match(code)


class prevent_embedded_answer(AssertionFeedback):
    """ Detects if the given code is present. """
    title = "Don't Write Answer Directly"
    message_template = ("You have embedded the answer directly in your code, "
                        "instead of writing code to compute the answer.")

    def __init__(self, code, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'code': code, 'root': root}
        super().__init__(fields=fields, **kwargs)

    def condition(self):
        """ Traverse the AST to find matches. """
        code = self.fields['code']
        ast = self.fields['root']
        return ast.find_match(code)


class prevent_printing_functions(AssertionFeedback):
    """
    Detects if the user has defined a function with a print statement inside.

    Args:
        exceptions: A set of function names (or a single string) of functions to allow to print (e.g., `'main'`).
    """
    title = "Do Not Print in Function"
    message_template = ("The function {name} is printing on line {location.line:line}."
                        " However, that function is not supposed to print.")

    def __init__(self, exceptions=None, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'root': root, 'exceptions': exceptions}
        super().__init__(fields=fields, **kwargs)

    def condition(self, *args, **kwargs):
        exceptions = self.fields['exceptions']
        if exceptions is None:
            exceptions = set()
        elif isinstance(exceptions, str):
            exceptions = {exceptions}
        root = self.fields['root']
        defs = root.find_all("FunctionDef")
        for a_def in defs:
            name = a_def._name
            if name in exceptions:
                continue
            for call in find_function_calls("print", root=a_def, report=self.report):
                location = call.locate()
                self.fields['name'] = name
                self.update_location(location)
                return True
        return False


class ensure_functions_return(AssertionFeedback):
    """
    Detects if the user has defined a function with a print statement inside.

    Args:
        exceptions: A set of function names (or a single string) of functions to allow to print (e.g., `'main'`).
    """
    title = "Must Return in Function"
    message_template = ("The function {name} is not returning."
                        " However, that function is supposed to have a return statement.")

    def __init__(self, exceptions=None, root=None, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        root = root or parse_program(report=report)
        fields = {'root': root, 'exceptions': exceptions}
        super().__init__(fields=fields, **kwargs)

    def condition(self, *args, **kwargs):
        exceptions = self.fields['exceptions']
        if exceptions is None:
            exceptions = set()
        elif isinstance(exceptions, str):
            exceptions = {exceptions}
        root = self.fields['root']
        defs = root.find_all("FunctionDef")
        for a_def in defs:
            name = a_def._name
            if name in exceptions:
                continue
            if not a_def.find_match("return"):
                self.fields['name'] = name
                return True
        return False


class only_printing_variables(AssertionFeedback):
    """
    Determines whether the user is only printing variables, as opposed to
    literal values.

    """
    title="Print Variables, Not Values"
    message_template=("You printed something other than a variable on"
                      " line {location.line:line}. Although that is not a"
                      " normally an issue, we want you to practice printing"
                      " variables in this problem.")

    def __init__(self, root=None, **kwargs):
        fields = kwargs.setdefault('fields', {})
        report = kwargs.setdefault('report', MAIN_REPORT)
        fields['root'] = root or parse_program(report=report)
        super().__init__(**kwargs)

    def condition(self):
        """ Uses find_all to detect matches, ignoring named values. """
        print_calls = find_function_calls('print', root=self.fields['root'],
                                          report=self.report)
        for print_call in print_calls:
            for arg in print_call.args:
                if arg.ast_name != "Name":
                    self.update_location(print_call.lineno)
                    return True
                elif arg.id in ("True", "False", "None"):
                    self.update_location(print_call.lineno)
                    return True
        return False


ADVANCED_ITERATION_FUNCTIONS = [
    "sum", "map", "filter", "reduce",
    "len", "max", "min", "max",
    "sorted", "all", "any",
    "getattr", "setattr",
    "eval", "exec", "iter", "next"
]


@CompositeFeedbackFunction(prevent_ast, prevent_function_call)
def prevent_advanced_iteration(allow_while=False, allow_for=False,
                               allow_function=None, **kwargs):
    """ Prevents the student from using certain advanced iteration functions
    and constructs. Does not currently support blocking recursion. """
    if isinstance(allow_function, str):
        allow_function = {allow_function}
    elif allow_function is None:
        allow_function = set()
    if not allow_while:
        prevent_ast("While")
    if not allow_for:
        prevent_ast("For")
    for function_name in ADVANCED_ITERATION_FUNCTIONS:
        if function_name not in allow_function:
            prevent_function_call(function_name, **kwargs)


class open_without_arguments(FeedbackResponse):
    """
    Detects if the user has called the `open` function without any arguments.
    """
    muted = False
    title = "Opened Without Arguments"
    message_template = ("You have called the `open` function "
                        "without any arguments. It needs a filename.")
    category = Feedback.CATEGORIES.INSTRUCTOR


@CompositeFeedbackFunction(open_without_arguments, ensure_literal)
def files_not_handled_correctly(*filenames, muted=False):
    """
    Statically detect if files have been opened and closed correctly.
    This is only useful in the case of very simplistic file handling.

    TODO: This could be vastly improved with line numbers and other information.
    """
    if filenames and isinstance(filenames[0], int):
        num_filenames = filenames[0]
        actual_filenames = False
    else:
        num_filenames = len(filenames)
        actual_filenames = True
    ast = parse_program()
    calls = ast.find_all("Call")
    called_open = []
    closed = []
    for a_call in calls:
        if a_call.func.ast_name == 'Name':
            if a_call.func.id == 'open':
                if not a_call.args:
                    open_without_arguments(muted)
                    return True
                called_open.append(a_call)
            elif a_call.func.id == 'close':
                explain("You have attempted to call `close` as a "
                        "function, but it is actually a method of the "
                        "file object.", label="used_close_as_function", title="Close Is a Method", priority='syntax')
                return True
        elif a_call.func.ast_name == 'Attribute':
            if a_call.func.attr == 'open':
                # TODO: Make these feedback functions
                explain("You have attempted to call `open` as a "
                        "method, but it is actually a built-in function.", label="used_open_as_method",
                        title="Open Is a Function")
                return True
            elif a_call.func.attr == 'close':
                closed.append(a_call)
    if len(called_open) < num_filenames:
        explain("You have not opened all the files you were supposed to.", label="unopened_files", title="Unopened Files")
        return True
    elif len(called_open) > num_filenames:
        explain("You have opened more files than you were supposed to.", label="extra_open_files", title="Extra Opened Files")
        return True
    withs = ast.find_all("With")
    if len(withs) + len(closed) < num_filenames:
        explain("You have not closed all the files you were supposed to.", label="unclosed_files", title="Unclosed Files")
        return True
    elif len(withs) + len(closed) > num_filenames:
        explain("You have closed more files than you were supposed to.", label="extra_closed_files", title="Extra Closed Files")
        return True
    if actual_filenames:
        for filename in filenames:
            ensured = ensure_literal(filename)
            if ensured:
                return ensured
    return False


"""
Utility function for taking various kinds of "type expressions" and turning
them into a canonical Pedal type.

* Pedal Type
* Python builtin type (e.g., `list`, `int`)
    get_pedal_type_from_builtin
* Python literal value (e.g., `5`, `"Hello World!"`)
    get_pedal_type_from_value
* AST Nodes (e.g., `ast.Num`)
    get_pedal_type_from_ast
* JSON-Encoded representation of a type (e.g., from CORGIS datasets)
    get_pedal_type_from_json
* String representation of parseable Python code
    get_pedal_type_from_str
* Python Typing module type (e.g., `List`, `Generic`)
    TODO: Implement get_pedal_type_from_typing
"""
import ast

from pedal.types.new_types import (Type, AnyType, ImpossibleType, NumType, BoolType,
                                   TupleType, ListType, StrType,
                                   DictType, SetType, GeneratorType,
                                   ModuleType, NoneType, FrozenSetType,
                                   FunctionType, TYPE_STRINGS,
                                   LiteralStr, LiteralInt, LiteralFloat, LiteralBool, ClassType, InstanceType,
                                   widest_type, LiteralValue, TypeUnion, PEDAL_TYPE_NAMES, specify_subtype)
from pedal.utilities.system import IS_AT_LEAST_PYTHON_39, IS_SKULPT

import types as dynamic_python_types

try:
    from dataclasses import fields
except:
    fields = None


def get_generic_type(type_expression, evaluate_name=None) -> Type:
    if not hasattr(type_expression, '__args__'):
        return TYPE_STRINGS[type_expression.__name__]()
    base_type = TYPE_STRINGS[type_expression.__name__]()
    arg_type = type_expression.__args__
    arg_type = tuple((normalize_type(at, evaluate_name) for at in arg_type))
    base_type.add_type_arguments(arg_type)
    return base_type


def unbox_sandbox_if_needed(value):
    """ TODO: Redefined here to avoid imports. Need a better solution! """
    if hasattr(value, "__actual_class__"):
        if value.__actual_class__.__name__ == 'SandboxResult':
            return value._actual_value
    return value


def normalize_type(type_expression, evaluate_name=None) -> Type:
    """
    Converts the given ``type_expression`` into a normalized Pedal Type.

    Args:
        type_expression (Any): Any kind of supported "Type Expression". This
            includes builtin Python types, string representations, AST nodes,
            JSON-Encoded representations.
        evaluate_name (function): A function that takes a string and returns
            a PedalType. This is used to evaluate strings into types, possibly via the current
            student runtime environment.

    Returns:
        :py:class:`pedal.types.new_types.Type`: A normalized Pedal Type.
    """
    # Already a Pedal Type
    if isinstance(type_expression, str) and type_expression in PEDAL_TYPE_NAMES:
        return PEDAL_TYPE_NAMES[type_expression]()
    if isinstance(type_expression, Type):
        return type_expression
    # What if it's a builtin type function?
    if isinstance(type_expression, type) or (IS_AT_LEAST_PYTHON_39 and isinstance(type_expression, dynamic_python_types.GenericAlias)):
        if type_expression.__name__ in TYPE_STRINGS:
            return get_generic_type(type_expression, evaluate_name)
        if evaluate_name:
            type_object = unbox_sandbox_if_needed(evaluate_name(type_expression.__name__))
            if isinstance(type_object, type) or (IS_AT_LEAST_PYTHON_39 and isinstance(type_object, dynamic_python_types.GenericAlias)):
                if fields and hasattr(type_object, '__dataclass_fields__'):
                    return ClassType(type_object.__name__, {
                        field.name: normalize_type(field.type, evaluate_name)
                        for field in fields(type_object)
                    })
                # TODO: Fix this ugly ugly hack!
                try:
                    vars(type_object)
                    has_vars = True
                except:
                    has_vars = False
                if has_vars:
                    return ClassType(type_object.__name__, {
                        f: normalize_type(v, evaluate_name)
                        for f, v in vars(type_object).items()
                        if not f.startswith('___')
                    })
                return ClassType(type_object.__name__, {})
            return type_object
        return normalize_type(type_expression.__name__, evaluate_name=evaluate_name)
    # Might be a string, can we evaluate?
    if isinstance(type_expression, str):
        # return get_pedal_type_from_str(type_expression, evaluate_name)
        try:
            first_line = ast.parse(type_expression).body[0].value
        except SyntaxError as e:
            raise ValueError(f"Could not parse string {type_expression!r} into "
                             f"a Pedal type:\n {e}")
        except AttributeError as e:
            raise ValueError(f"Could not parse string {type_expression!r} into "
                             f"a Pedal type:\n This was not an expression.")
        try:
            return normalize_type(first_line, evaluate_name=evaluate_name)
        except:
            return ImpossibleType()
    # Might be a JSON-Encoded dictionary
    if isinstance(type_expression, dict):
        try:
            return get_pedal_type_from_json(type_expression, evaluate_name)
        except KeyError:
            return get_pedal_type_from_type_literal(type_expression, evaluate_name)
    # Perhaps it is an AST node
    if isinstance(type_expression, ast.AST):
        return get_pedal_type_from_ast(type_expression, evaluate_name)
    if isinstance(type_expression, (set, list, tuple, frozenset)):
        return get_pedal_type_from_type_literal(type_expression, evaluate_name)
    if isinstance(type_expression, dynamic_python_types.ModuleType):
        return ModuleType(type_expression.__name__, {
            field: normalize_type(getattr(type_expression, field), evaluate_name)
            for field in dir(type_expression)
            if not field.startswith('__')
        })
    return get_pedal_type_from_value(type_expression)
    # raise ValueError(f"Could not normalize {type_expression!r} into a Pedal type.")


def get_pedal_type_from_json(val, evaluate_name=None):
    if 'version' in val:
        if val['version'] == 2:
            return get_pedal_type_from_json_v2(val, evaluate_name)
    return get_pedal_type_from_json_v1(val, evaluate_name)


def get_pedal_type_from_json_v2(val, evaluate_name=None):
    """ TODO: Finish v2 to make this cleaner """
    return get_pedal_type_from_json_v1(val, evaluate_name)


def get_pedal_type_from_json_v1(val, evaluate_name=None):
    if val['type'] == 'LiteralStr':
        return LiteralStr(val['value'])
    elif val['type'] == 'LiteralNum':
        if isinstance(val['value'], int):
            return LiteralInt(val['value'])
        else:
            return LiteralFloat(val['value'])
    elif val['type'] == 'LiteralBool':
        return LiteralBool(val['value'])
    elif val['type'] == 'DictType':
        values = [get_pedal_type_from_json_v1(v, evaluate_name)
                  for v in val['values']]
        if 'literals' in val:
            keys = [get_pedal_type_from_json_v1(l, evaluate_name)
                    for l in val['literals']]
        else:
            keys = [get_pedal_type_from_json_v1(k, evaluate_name)
                    for k in val['keys']]
        return DictType([(k, v) for k, v in zip(keys, values)])
    elif val['type'] == 'ListType':
        return ListType(val.get('empty', None),
                        get_pedal_type_from_json_v1(val.get('subtype', None),
                                                    evaluate_name))
    elif val['type'] == 'StrType':
        return StrType(val.get('empty', None))
    elif val['type'] == 'BoolType':
        return BoolType()
    elif val['type'] == 'NoneType':
        return NoneType()
    elif val['type'] == 'NumType':
        return NumType()
    elif val['type'] == 'ModuleType':
        submodules = {name: get_pedal_type_from_json_v1(m, evaluate_name)
                      for name, m in val.get('submodules', {}).items()}
        fields = {name: get_pedal_type_from_json_v1(m, evaluate_name)
                  for name, m in val.get('fields', {}).items()}
        return ModuleType(name=val.get('name'), submodules=submodules,
                          fields=fields)
    elif val['type'] == 'FunctionType':
        returns = get_pedal_type_from_json_v1(val.get('returns', {'type': 'NoneType'}),
                                              evaluate_name)
        return FunctionType(name=val.get('name'), returns=lambda: returns)


def get_pedal_type_from_str(value, evaluate_name=None):
    """

    Args:
        value:
        tifa_instance:

    Returns:

    """
    # if value in custom_types:
    #    return custom_types[value]
    if value.lower() in TYPE_STRINGS:
        return TYPE_STRINGS[value.lower()]()
    elif evaluate_name:
        potential = normalize_type(evaluate_name(value), evaluate_name)
        if not isinstance(potential, ImpossibleType):
            return potential
    return ClassType(value)
    # TODO: handle custom types


ELEMENT_TYPES = {
    set: SetType, list: ListType, frozenset: FrozenSetType
}


def get_pedal_type_from_value(value, evaluate_name=None) -> Type:
    """ Converts the Python value to a Pedal Type """
    if isinstance(value, bool):
        return LiteralBool(value)
    if isinstance(value, int):
        return LiteralInt(value)
    if isinstance(value, float):
        return LiteralFloat(value)
    if isinstance(value, complex):
        return NumType()
    if isinstance(value, str):
        return LiteralStr(value)
    if isinstance(value, type(None)):
        return NoneType()
    if isinstance(value, tuple):
        return TupleType((get_pedal_type_from_value(t, evaluate_name) for t in value))
    if isinstance(value, (list, set, frozenset)):
        container_type = ELEMENT_TYPES.get(type(value), ListType)
        if value:
            element_type = widest_type([get_pedal_type_from_value(v, evaluate_name)
                                        for v in value])
            if element_type is not None:
                return container_type(False, element_type)
            else:
                return container_type(False, get_pedal_type_from_value(value[0], evaluate_name))
        else:
            return container_type(True)
    if isinstance(value, dict):
        if not value:
            return DictType([])
        # All literal keys
        items = [(get_pedal_type_from_value(key, evaluate_name),
                  get_pedal_type_from_value(value, evaluate_name))
                 for key, value in value.items()]
        if all(isinstance(k, LiteralValue) for k, _ in items):
            return DictType(list(items))
        # Find the widest possible single type
        first_key = widest_type([k for k, _ in items])
        first_value = widest_type([v for _, v in items])
        if first_key is not None and first_value is not None:
            return DictType([(first_key, first_value)])
        else:
            # Resort to just matching the types?
            return DictType([(k, v) for k, v in items])
    if evaluate_name:
        new_instance = InstanceType(normalize_type(evaluate_name(type(value).__name__), evaluate_name))
        if fields and hasattr(value, '__dataclass_fields__'):
            for field in fields(value):
                new_instance.add_attr(field.name, get_pedal_type_from_value(getattr(value, field.name), evaluate_name))
        return new_instance
    return AnyType()


def get_pedal_type_from_ast(value: ast.AST, evaluate_name=None) -> Type:
    """
    Determines the Pedal Type from this ast node.
    Args:
        value (ast.AST): An AST node.

    Returns:
        Type: A Pedal Type
    """
    try:
        if isinstance(value, ast.Constant):
            return get_pedal_type_from_value(value.value, evaluate_name)
    except AttributeError as e:
        pass
    if isinstance(value, ast.Name):
        return get_pedal_type_from_str(value.id, evaluate_name)
    elif isinstance(value, ast.Str):
        return LiteralStr(value.s)
    elif isinstance(value, ast.List):
        return ListType(not bool(value.elts),
                        (get_pedal_type_from_ast(value.elts[0], evaluate_name)
                         if value.elts else None))
    elif isinstance(value, ast.Set):
        return SetType(not bool(value.elts),
                       (get_pedal_type_from_ast(value.elts[0], evaluate_name)
                        if value.elts else None))
    elif isinstance(value, ast.Tuple):
        return TupleType([get_pedal_type_from_ast(e, evaluate_name)
                           for e in value.elts])
    elif isinstance(value, ast.Dict):
        if not value.keys:
            return DictType([])
        return DictType([(get_pedal_type_from_ast(k, evaluate_name),
                          get_pedal_type_from_ast(v, evaluate_name))
                         for k, v in zip(value.keys, value.values)])
    # Support new style subscripts (e.g., ``list[int]``)
    elif ((IS_AT_LEAST_PYTHON_39 and isinstance(value, ast.Subscript)) or
          isinstance(value, ast.Subscript) and isinstance(value.slice, ast.Index)):
        if IS_AT_LEAST_PYTHON_39 and not IS_SKULPT:
            slice = value.slice
        else:
            slice = value.slice.value
        slice_type = get_pedal_type_from_ast(slice, evaluate_name)
        value_type = get_pedal_type_from_ast(value.value, evaluate_name)
        result = value_type.index(slice_type)
        if isinstance(slice, ast.Slice):
            return value_type.shallow_clone()
        else:
            return result
    # Top-level Module, parse it and get it back
    if isinstance(value, ast.Module) and value.body:
        if isinstance(value.body[0], ast.Expr):
            return get_pedal_type_from_ast(value.body[0].value, evaluate_name)
    if isinstance(value, ast.Expr):
        return get_pedal_type_from_ast(value.value, evaluate_name)
    return ImpossibleType()


def get_pedal_type_from_type_literal(value, evaluate_name=None) -> Type:
    if value == list:
        # Handle type annotations if available
        return ListType(False)
    elif value == dict:
        return DictType(False)
    elif isinstance(value, list):
        if value:
            return ListType(False, normalize_type(value[0], evaluate_name))
        else:
            return ListType(True)
    elif isinstance(value, (set, frozenset)):
        return TypeUnion([normalize_type(v, evaluate_name) for v in value])
    elif isinstance(value, tuple):
        return TupleType([normalize_type(v, evaluate_name) for v in value])
    elif isinstance(value, dict):
        if not value:
            return DictType(True)
        if all(isinstance(value, str) for k in value.keys()):
            return DictType([(LiteralStr(s), normalize_type(vv, evaluate_name))
                            for s, vv in value.items()])
        return DictType([(normalize_type(kk, evaluate_name),
                          normalize_type(vv, evaluate_name))
                         for kk, vv in value.items()])



from pedal.types.new_types import BUILTIN_MODULES, ModuleType, FunctionType, NoneType, register_builtin_module

register_builtin_module('cisc108', lambda: ModuleType('cisc108', fields={
    'assert_equal': FunctionType(name='assert_equal', returns=NoneType),
    'assert_type': FunctionType(name='assert_type', returns=NoneType),
}))

register_builtin_module('cisc106', lambda: ModuleType('cisc106', fields={
    'assert_equal': FunctionType(name='assert_equal', returns=NoneType),
    'assert_type': FunctionType(name='assert_type', returns=NoneType),
}))

register_builtin_module('bakery', lambda: ModuleType('bakery', fields={
    'assert_equal': FunctionType(name='assert_equal', returns=NoneType),
    'assert_type': FunctionType(name='assert_type', returns=NoneType),
}))


from pedal.types.new_types import ModuleType, FunctionType, void_function, ClassType, \
    InstanceType, bool_function, int_function, str_function, register_builtin_module

_VOID_FUNCTIONS = ['draw', 'start', 'debug', 'stop', 'stop_music', 'set_window_size', 'background_image',
                   'background_music', 'continue_music', 'destroy', 'pause', 'pause_music', 'play_music',
                   'play_sound', 'rewind_music', 'set_keyboard_delay', 'set_game_state',
                   'set_keyboard_interval', 'set_keyboard_repeat', 'set_mouse_cursor', 'set_mouse_position',
                   'set_mouse_visible', 'set_music_position', 'set_music_volume', 'set_window_color',
                   'set_window_size', 'set_window_state', 'set_window_title', 'set_world_state',
                   'disable_keyboard_repeating', 'enable_keyboard_repeating', 'when', 'unregister']
_DESIGNER_FUNCTIONS = ['circle', 'rectangle', 'arc', 'emoji', 'group', 'image',
                       'lines', 'pen', 'line', 'shape', 'text',
                       'Arc', 'DesignerObject', 'designer_object', 'ellipse']
_IDENTITY_FUNCTIONS = ['above', 'below', 'beside', 'spin',
                       'move_forward', 'move_backward', 'turn_left', 'turn_right', 'go_to', 'go_to_xy', 'go_to_mouse',
                       'point_towards', 'point_towards_mouse', 'point_in_direction', 'change_xy', 'change_x',
                       'change_y', 'set_x', 'set_y',
                       'get_angle', 'get_x', 'get_y',
                       'flip_x', 'flip_y', 'set_flip_x', 'set_flip_y', 'set_scale', 'set_scale_x', 'set_scale_y',
                       'set_background_image',
                       'get_scale', 'get_scale_x', 'get_scale_y', 'get_visible', 'get_flip_x', 'get_flip_y', 'show',
                       'hide',
                       'grow', 'grow_x', 'grow_y', 'shrink',
                       'move_to_x', 'move_to_y', 'move_to', 'move_to_mouse', 'move_to_xy',
                       'set_visible', 'change_scale', 'set_emoji_name', 'set_text']


def build_designer_module():
    _DesignerObject = ClassType('DesignerObject', {}, [])

    def _DesignerObject_constructor():
        return InstanceType(_DesignerObject)

    _DESIGNER_FIELDS = {
        'colliding': bool_function('colliding'),
        'colliding_with_mouse': bool_function('colliding_with_mouse'),
        'get_window_title': str_function('get_window_title'),
        'get_mouse_x': int_function('get_mouse_x'),
        'get_mouse_y': int_function('get_mouse_y'),
        'get_height': int_function('get_height'),
        'get_width': int_function('get_width'),
        'get_keyboard_delay': int_function('get_keyboard_delay'),
        'get_keyboard_interval': int_function('get_keyboard_interval'),
        'get_keyboard_repeat': bool_function('get_keyboard_repeat'),
        'get_emoji_name': str_function('get_emoji_name'),
        'get_text': str_function('get_text'),
    }

    _DESIGNER_FIELDS.update({
        name: void_function(name) for name in _VOID_FUNCTIONS
    })
    _DESIGNER_FIELDS.update({
        name: FunctionType(name, returns=_DesignerObject_constructor)
        for name in _DESIGNER_FUNCTIONS
    })
    _DESIGNER_FIELDS.update({
        name: FunctionType(name, returns='identity')
        for name in _IDENTITY_FUNCTIONS
    })
    return ModuleType('designer', _DESIGNER_FIELDS)


register_builtin_module('designer', build_designer_module)



"""
Outputs in GradeScope format

Dumps the JSON in the expected file location.

$> pedal feedback \
        /autograder/run_autograder/on_run.py \
        /autograder/submission \
        --environment gradescope > \
        "/autograder/results/results.json"

{ "score": 44.0, // optional, but required if not on each test case below. Overrides total of tests if specified.
  "output": "Text relevant to the entire submission", // optional
  "visibility": "after_due_date", // Optional visibility setting
  "stdout_visibility": "visible", // Optional stdout visibility setting
  "extra_data": {}, // Optional extra data to be stored
  "tests": // Optional, but required if no top-level score
    [
        {
            "score": 2.0, // optional, but required if not on top level submission
            "max_score": 2.0, // optional
            "name": "Your name here", // optional
            "number": "1.1", // optional (will just be numbered in order of array if no number given)
            "output": "Giant multiline string that will be placed in a <pre> tag and collapsed by default", // optional
            "tags": ["tag1", "tag2", "tag3"], // optional
            "visibility": "visible", // Optional visibility setting
            "extra_data": {} // Optional extra data to be stored
        },
        // and more test cases...
    ],
  "leaderboard": // Optional, will set up leaderboards for these values
    [
      {"name": "Accuracy", "value": .926},
      {"name": "Time", "value": 15.1, "order": "asc"},
      {"name": "Stars", "value": "*****"}
    ]
}

"""
import json
import tabulate

from pedal import Score
from pedal.core.final_feedback import parse_feedback
from pedal.resolvers.core import make_resolver
from pedal.source import verify, next_section as original_next_section
from pedal.core.environment import Environment
from pedal.core.report import MAIN_REPORT
from pedal.sandbox import run, get_sandbox, set_input, start_trace, get_output
from pedal.source.sections import FeedbackSourceSection
from pedal.tifa import tifa_analysis
from pedal.resolvers.simple import resolve as simple_resolve, by_priority
from pedal.resolvers.full import resolve as full_resolve
from pedal.resolvers.sectional import resolve as original_sectional_resolve
from pedal.core.formatting import Formatter, HtmlFormatter


class GradeScopeEnvironment(Environment):
    """
    Configures the GradeScope programming environment.

    # TODO: Add command line argument to use "simple" resolver instead of full
    """
    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='on_run.py', skip_tifa=False, skip_run=False,
                 inputs=None, set_correct=True, set_success=None,
                 report=MAIN_REPORT, trace=True, threaded=True, **kwargs):
        super().__init__(files=files, main_file=main_file, main_code=main_code,
                         user=user, assignment=assignment, course=course,
                         execution=execution, instructor_file=instructor_file,
                         report=report)
        self.skip_run = skip_run
        self.skip_tifa = skip_tifa
        self.trace = trace
        report.set_formatter(Formatter(report))
        verify(report=self.report)
        if not skip_tifa:
            tifa_analysis(report=self.report)
        if inputs:
            set_input(inputs)
        if skip_run:
            student = get_sandbox(report=report)
            student.threaded = threaded
        else:
            if trace:
                start_trace()
            student = run(report=report, threaded=threaded)
            student.threaded = threaded
        self.fields = {
            'student': student,
            'resolve': resolve,
            'next_section': self.next_section,
            'sectional_resolve': sectional_resolve,
            'simple': single_resolve,
            'single': single_resolve,
            'show_as_hidden': show_as_hidden
        }

    def next_section(self, name=""):
        original_next_section(name=name, report=self.report)
        verify(report=self.report)
        if not self.skip_tifa:
            tifa_analysis(report=self.report)
        student = get_sandbox(report=self.report)
        if self.skip_run:
            student.clear()
        else:
            if self.trace:
                start_trace()
            student.clear()
            student = student.run()
        return student

    def load_main(self, path):
        """
        Allowed to return either a string value (the contents of the file)
        or the exception that was raised.
        """
        try:
            return super().load_main(path)
        except OSError as file_error:
            return file_error


setup_environment = GradeScopeEnvironment


def dump_feedback(**results):
    print(json.dumps(results, indent=2))


score_maximum = 1


@make_resolver
def detailed_resolver(report=MAIN_REPORT, priority_key=by_priority):
    """ Group by category, provide all the non-muted activated feedback """
    pass


@make_resolver
def resolve(report=MAIN_REPORT, customize_failure_message= None, leaderboard=None):
    """

    TODO: Support some hidden tests for instructors

    Args:
        leaderboard: A list of data for the leaderboard, or a function to generate a leaderboard.
        customize_failure_message: A function that takes in the final feedback object in order to manipulate its
            settings. This is useful for changing the message of a feedback object to be more specific to the
            needs of the instructor.
        report: The report to resolve.
    """
    final = full_resolve(report)
    tests = []
    for feedback in final.used:
        test = {"name": feedback.title}
        message = feedback.message if feedback else (feedback.else_message or "Correct")
        if feedback.title != message:
            test['output'] = message
        # Handle scoring
        success, partial, message, title, data = parse_feedback(feedback)
        if not feedback.unscored and feedback.score is not None:
            # Triggered Negative leads to opposite behavior for operator
            # Also untriggered positive feedback
            invert_logic = ((feedback.valence != feedback.NEGATIVE_VALENCE) == (not feedback))
            inversion = "!" if invert_logic else ""
            score = Score.parse(f"{inversion}{partial}")
            if not score.invert:
                test['score'] = score.add_to_current(0) * score_maximum
            else:
                test['score'] = 0
        # But do we actually show it?
        if getattr(feedback, 'gradescope_visibility', 'visible') == 'hidden':
            test['visibility'] = 'hidden'
            tests.append(test)
        elif feedback.parent is None:
            tests.append(test)

    if not final.success and final.score < 1:
        if customize_failure_message is not None:
            customize_failure_message(final)
        else:
            final.title = "Failed Instructor Tests"
            final.message = "Check the results below to see what you have failed."
    extra = {}
    if isinstance(leaderboard, list):
        extra['leaderboard'] = leaderboard
    elif leaderboard is not None:
        extra['leaderboard'] = leaderboard(final)

    dump_feedback(
        score=round(final.score*score_maximum, 2),
        output=f"<strong>{final.title}</strong><br>\n{final.message}",
        tests=tests,
        **extra
    )
    return final


def single_resolve(report=MAIN_REPORT, leaderboard=None):
    final = full_resolve(report)
    extra = {}
    if isinstance(leaderboard, list):
        extra['leaderboard'] = leaderboard
    elif leaderboard is not None:
        extra['leaderboard'] = leaderboard(final)

    dump_feedback(
        score=round(final.score * score_maximum, 2),
        output=f"<strong>{final.title}</strong><br>\n{final.message}",
        **extra
    )
    return final


def show_as_hidden(feedback):
    feedback.gradescope_visibility = 'hidden'


@make_resolver
def sectional_resolve(report=MAIN_REPORT, priority_key=by_priority):
    """

    Args:
        report:
        custom_success_message:
    """
    finals = original_sectional_resolve(report, priority_key=priority_key)
    #if final.positives:
    #    print("-Positive Notes")
    #    for positive in final.positives:
    #        print(positive)
    global_feedbacks = sorted([f for f in report.feedback if f.parent is None],
                              key=lambda f: f.section_number if isinstance(f, FeedbackSourceSection) else -1)
    tests = []
    for global_feedback in global_feedbacks:
        if isinstance(global_feedback, FeedbackSourceSection):
            test = {"number": global_feedback.section_number,
                    "name": f"Part {global_feedback.section_number}"}
            final = finals.get(global_feedback)
            if final:
                test['output'] = final.title + "\n" + final.message
            else:
                test['output'] = "No feedback for this section"
        # TODO: Should we be skipping all global feedback? Need more examples
    total_score = sum(f.score for f in finals.values())
    dump_feedback(
        score=round(total_score * score_maximum),
        output="Feedback is below",
        tests=tests
    )
    return list(finals.values())[0]


def set_maximum_score(number):
    """
    TODO: Attach data to the Report instead of a global variable.

    Args:
        number:
        cap:
        report:
    """
    global score_maximum
    score_maximum = number


class GradeScopeFormatter(HtmlFormatter):

    def python_code(self, code):
        return self.pre(code)

    def python_expression(self, code):
        return self.html_code(code)

    def filename(self, filename):
        return filename

    def python_value(self, code):
        return self.pre(code)

    def inputs(self, inputs):
        return self.pre(inputs)

    def output(self, output):
        return self.pre(output)

    def traceback(self, traceback):
        return traceback

    def name(self, name):
        return name

    def line(self, line_number):
        return line_number

    def frame(self, name):
        return name

    def exception(self, exception):
        return self.pre(exception)

    def table(self, rows, columns):
        return self.pre(tabulate.tabulate(rows, headers=columns))
